import argparse
import json
import multiprocessing as mp

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from pymysql.cursors import SSCursor

from immunedb.common.models import Base


def get_base_arg_parser(desc='', multiproc=True, **kwargs):
    """Gets a base argument parser which requires database configuration.

    :param str desc: The description provided by the argument parser

    :return: The argument parser object
    :rtype: ArgumentParser

    """
    parser = argparse.ArgumentParser(
            description=desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            **kwargs)

    parser.add_argument('db_config', help='Path to database config')
    if multiproc:
        try:
            num_cpu = mp.cpu_count()
        except NotImplementedError:
            num_cpu = 4
        parser.add_argument('--nproc', default=num_cpu, type=int,
                            help='Number of subprocesses to run')

    return parser


def init_db(database_config, drop_all=False, as_maker=False, create=True):
    """Initializes a session with the specified database.

    :param str database_config: If ``from_dict`` is ``False``, the path to
        data database config file, otherwise a database config dictionary
    :param bool as_maker: If ``True``, the returned object will be a session
        maker rather than an session

    :returns: A ``session`` or, if ``as_maker`` is set, a ``session_maker``

    """

    if isinstance(database_config, basestring):
        with open(database_config) as fh:
            database_config = json.load(fh)

    conn = 'mysql+pymysql://{}:{}@{}/{}?charset=utf8&use_unicode=0'.format(
        database_config['username'], database_config['password'],
        database_config['host'], database_config['database'])
    engine = create_engine(conn, pool_recycle=3600,
                           connect_args={'cursorclass': SSCursor})

    if drop_all:
        Base.metadata.drop_all(engine)

    if create:
        Base.metadata.create_all(engine)

    session = sessionmaker()
    session.configure(bind=engine)

    return session if as_maker else session()



from collections import OrderedDict

from immunedb.common.models import Sequence, SequenceCollapse
from immunedb.exporting.export import Exporter


def _if_clone(f):
    def _wrap(seq):
        if seq.clone_id is not None:
            return f(seq)
        return ''
    return _wrap


class SequenceExport(Exporter):
    """A class to handle exporting sequences in various formats.

    :param Session session: A handle to a database session
    :param str eformat: The export format to use.  Currently "csv", "orig", and
        "clip" for tab-delimited, FASTA, FASTA with filled in germlines, and
        FASTA in CLIP format respectively
    :param str rtype: The type of record to filter the query on.  Currently
        either "sample" or "clone"
    :param int-array rids: A list of IDs of ``rtype`` to export
    :param str-array selected_fields: A list of fields to export
    :param bool noresults: If sequences which could not be assigned a V or J
        should be included in the output

    """

    allowed_fields = OrderedDict({
        'seq_id': lambda seq: seq.seq_id,
        'subject_id': lambda seq: seq.sample.subject.id,
        'subject_identifier': lambda seq: seq.sample.subject.identifier,

        'sample_id': lambda seq: seq.sample_id,
        'sample_name': lambda seq: seq.sample.name,
        'study_id': lambda seq: seq.sample.study.id,
        'study_name': lambda seq: seq.sample.study.name,

        'partial': lambda seq: seq.partial,
        'probable_indel_or_misalign':
            lambda seq: seq.probable_indel_or_misalign,
        'num_gaps': lambda seq: seq.num_gaps,
        'seq_start': lambda seq: seq.seq_start,
        'v_match': lambda seq: seq.v_match,
        'v_length': lambda seq: seq.v_length,
        'j_match': lambda seq: seq.j_match,
        'j_length': lambda seq: seq.j_length,
        'pre_cdr3_match': lambda seq: seq.pre_cdr3_match,
        'pre_cdr3_length': lambda seq: seq.pre_cdr3_length,
        'post_cdr3_match': lambda seq: seq.post_cdr3_match,
        'post_cdr3_length': lambda seq: seq.post_cdr3_length,
        'in_frame': lambda seq: seq.in_frame,
        'functional': lambda seq: seq.functional,
        'stop': lambda seq: seq.stop,
        'copy_number': lambda seq: seq.copy_number,
        'sequence': lambda seq: seq.sequence,
        'quality': lambda seq: seq.quality,
        'germline': lambda seq: seq.germline,
        'v_gene': lambda seq: seq.v_gene,
        'j_gene': lambda seq: seq.j_gene,
        'cdr3_nt': lambda seq: seq.cdr3_nt,
        'cdr3_aa': lambda seq: seq.cdr3_aa,
        'cdr3_num_nts': lambda seq: seq.cdr3_num_nts,
        'insertions': lambda seq: seq._insertions,
        'deletions': lambda seq: seq._deletions,

        'clone_id': _if_clone(lambda seq: seq.clone_id),
        'clone_cdr3_nt': _if_clone(lambda seq: seq.clone.cdr3_nt),
        'clone_cdr3_aa': _if_clone(lambda seq: seq.clone.cdr3_aa),
        'clone_json_tree': _if_clone(lambda seq: seq.clone.tree),
        'clone_germline': _if_clone(lambda seq: seq.clone.germline),
        'clone_parent_id': _if_clone(lambda seq: seq.clone.parent_id),

        'copy_number_in_subject':
            lambda seq: seq.collapse.copy_number_in_subject,
        'collapse_to_subject_sample_id': lambda seq:
            seq.collapse.collapse_to_subject_sample_id,
        'collapse_to_subject_seq_id': lambda seq:
            seq.collapse.collapse_to_subject_seq_id,
        'instances_in_subject': lambda seq:
            seq.collapse.instances_in_subject,
    })

    def __init__(self, session, writer, rtype, rids, selected_fields,
                 subject_uniques, only_with_clones):
        super(SequenceExport, self).__init__(
            session, rtype, rids, SequenceExport.allowed_fields,
            selected_fields)
        self.writer = writer
        self.subject_uniques = subject_uniques
        self.only_with_clones = only_with_clones

    def get_data(self):
        """Gets the output data for this export instance.  This could be export
        to a file, over a socket, etc. as it's simply a string or could be
        treated as a byte string.

        :returns: Lines of output for the selected sequences
        :rtype: str

        """

        # Get all the sequences matching the request
        self.writer.set_selected_fields(self.selected_fields)
        queries = [self.session.query(Sequence).filter(
            getattr(
                Sequence, '{}_id'.format(self.rtype)
            ) == rid
        ) for rid in sorted(self.rids)]
        for query in queries:
            query = query.join(SequenceCollapse)
            if self.subject_uniques:
                query = query.filter(
                    SequenceCollapse.copy_number_in_subject >= 1
                )
            if self.only_with_clones:
                query = query.filter(~Sequence.clone_id.is_(None))

            query = self.writer.preprocess_query(query)
            self.selected_fields = self.writer.get_required_fields(
                self.selected_fields)

            for seq in query:
                # Get the selected data for the sequence
                data = self.get_selected_data(seq)
                yield self.writer.format_sequence(data)



from collections import OrderedDict

from immunedb.common.models import CloneStats
from immunedb.exporting.export import Exporter
from immunedb.util.nested_writer import NestedCSVWriter


def _if_sample(f, val=''):
    def _wrap(stat):
        if stat.sample_id is not None:
            return f(stat)
        return val
    return _wrap


class CloneExport(Exporter):
    allowed_fields = OrderedDict({
        'clone_id': lambda s: s.clone_id,
        'sample_id': _if_sample(lambda s: s.sample_id, 'Total'),
        'unique_sequences': lambda s: s.unique_cnt,
        'total_sequences': lambda s: s.total_cnt,

        'v_gene': lambda s: s.clone.v_gene,
        'j_gene': lambda s: s.clone.j_gene,
        'cdr3_nt': lambda s: s.clone.cdr3_nt,
        'cdr3_aa': lambda s: s.clone.cdr3_aa,
        'cdr3_num_nts': lambda s: s.clone.cdr3_num_nts,
        'functional': lambda s: s.clone.functional,
        'insertions': lambda s: s.clone._insertions,
        'deletions': lambda s: s.clone._deletions,

        'sample_name': _if_sample(lambda s: s.sample.name, 'Total'),
        'subject_id': _if_sample(lambda s: s.sample.subject.id),
        'subject_identifier':
            _if_sample(lambda s: s.sample.subject.identifier),
        'tree': lambda s: s.clone.tree,
        'parent_id': lambda s: s.clone.parent_id
    })

    def __init__(self, session, rtype, rids, selected_fields):
        super(CloneExport, self).__init__(
            session, rtype, rids, CloneExport.allowed_fields,
            selected_fields)

    def get_data(self):
        csv = NestedCSVWriter(self.selected_fields, streaming=True)
        if self.rtype == 'sample':
            stats = self.session.query(
                CloneStats.clone_id
            ).filter(
                CloneStats.sample_id.in_(self.rids)
            )
            clone_ids = [e.clone_id for e in stats]
        else:
            clone_ids = self.rids

        for clone_id in sorted(set(clone_ids)):
            stats = self.session.query(CloneStats).filter(
                CloneStats.clone_id == clone_id
            ).order_by(CloneStats.sample_id)
            for stat in stats:
                yield csv.add_row(self.get_selected_data(stat))



import json
import shlex

from subprocess import Popen, PIPE

import ete2

from immunedb.common.models import Sequence, SequenceCollapse


class PhylogeneticTree(object):
    def __init__(self, germline_sequence, sequences, min_mut_occurrence=1,
                 min_mut_samples=1):
        self.germline_sequence = germline_sequence
        self.sequences = sequences
        self.min_mut_occurrence = min_mut_occurrence
        self.min_mut_samples = min_mut_samples

    def run(self, session, clearcut_path):
        fasta, removed_muts = get_fasta_input(
            self.germline_sequence, self.sequences,
            min_mut_occurrence=self.min_mut_occurrence,
            min_mut_samples=self.min_mut_samples)

        newick = get_newick(fasta, clearcut_path)
        tree = populate_tree(session, newick, self.germline_sequence,
                             removed_muts)
        tree.set_outgroup('germline')
        tree.search_nodes(name='germline')[0].delete()

        first = True
        while True:
            push_common_mutations_up(tree, first)
            remove_parent_mutations(tree)
            rem_null = remove_null_nodes(tree)
            moved = check_supersets(tree)
            if not rem_null and not moved and not are_null_nodes(tree):
                break
            first = False

        self.tree = tree
        return self.tree


def populate_tree(session, newick, germline_seq, removed_muts):
    tree = ete2.Tree(newick)
    for node in tree.traverse():
        if node.name not in ('NoName', 'germline', ''):
            seq = session.query(Sequence).filter(
                Sequence.ai == node.name
            ).first()
            seq_ids = {}
            for collapsed_seq in get_seqs_collapsed_to(session, seq):
                seq_ids[collapsed_seq.seq_id] = {
                    'ai': collapsed_seq.ai,
                    'copy_number': collapsed_seq.copy_number,
                    'sample_name': collapsed_seq.sample.name,
                    'sample_id': collapsed_seq.sample.id,
                    'metadata': collapsed_seq.sample.metadata_dict
                }

            node.name = seq.seq_id
            node.add_feature('seq_ids', seq_ids)
            node.add_feature('copy_number', sum(
                [s['copy_number'] for s in seq_ids.values()]
            ))
            modified_seq = remove_muts(seq.sequence,
                                       removed_muts, germline_seq)
            node.add_feature('mutations', get_mutations(
                germline_seq, modified_seq,
                map(int, json.loads(seq.mutations_from_clone).keys())))
        else:
            node = instantiate_node(node)

    return tree


def get_fasta_input(germline_seq, sequences, min_mut_occurrence=1,
                    min_mut_samples=1):
    seqs = {}
    mut_counts = {}
    for seq in sequences:
        seqs[seq.ai] = seq.clone_sequence
        if seq.mutations_from_clone is None:
            raise Exception(
                'Mutation information not available for sequence '
                '{}. Skipping this clone tree. Was '
                'immunedb_clone_stats run?'.format(seq.seq_id)
            )

        mutations = get_mutations(
            germline_seq, seq.clone_sequence, map(
                int, json.loads(seq.mutations_from_clone).keys())
        )
        for mut in mutations:
            if mut not in mut_counts:
                mut_counts[mut] = {'count': 0, 'samples': set([])}
            mut_counts[mut]['count'] += 1
            mut_counts[mut]['samples'].add(seq.sample_id)

    removed_muts = set([])
    for mut, cnts in mut_counts.iteritems():
        if (cnts['count'] < min_mut_occurrence or
                len(cnts['samples']) < min_mut_samples):
            removed_muts.add(mut)

    for seq_id, seq in seqs.iteritems():
        seqs[seq_id] = remove_muts(seq, removed_muts, germline_seq)

    in_data = '>germline\n{}\n'.format(germline_seq)
    for seq_id, seq in seqs.iteritems():
        in_data += '>{}\n{}\n'.format(seq_id, seq)
    return in_data, removed_muts


def get_seqs_collapsed_to(session, seq):
    sample_level_seqs = session.query(SequenceCollapse.seq_ai).filter(
        SequenceCollapse.collapse_to_subject_seq_ai == seq.ai
    )
    for sample_seq in sample_level_seqs:
        yield session.query(
            Sequence
        ).filter(Sequence.ai == sample_seq.seq_ai).one()


def remove_muts(seq, removes, germline_seq):
    for mut in removes:
        loc, _, to = mut
        loc -= 1
        if seq[loc] == to:
            seq = seq[:loc] + germline_seq[loc] + seq[loc + 1:]
    return seq


def get_newick(fasta_input, tree_prog):
    proc = Popen(shlex.split('{} --alignment -q --DNA -N -r'.format(
        tree_prog)), stdin=PIPE, stdout=PIPE, stderr=PIPE)
    return proc.communicate(input=fasta_input)[0]


def get_mutations(s1, s2, positions):
    return set([(i + 1, s1[i], s2[i]) for i in positions if s1[i] != s2[i]])


def instantiate_node(node):
    node.add_feature('seq_ids', {})
    node.add_feature('copy_number', 0)
    node.add_feature('sequence', None)
    node.add_feature('mutations', set([]))

    return node


def get_nested(seqs, key):
    return sorted(set(
        [s['metadata'][key] for s in seqs.values() if key in s['metadata']]
    ))


def tree_as_dict(tree, root=True):
    all_meta = set(
        [k for seq in tree.seq_ids.values() for k in seq['metadata'].keys()]
    )
    node = {
        'data': {
            'seq_ids': tree.seq_ids,
            'copy_number': tree.copy_number,
            'metadata': {
                k: get_nested(tree.seq_ids, k) for k in all_meta
            },
            'mutations': [{
                'pos': mut[0],
                'from': mut[1],
                'to': mut[2],
            } for mut in tree.mutations]
        },
        'children': []
    }
    for child in tree.children:
        node['children'].append(tree_as_dict(child, root=False))

    if not root or (len(tree.mutations) == 0 and len(tree.seq_ids) == 0):
        return node

    return {
        'data': {
            'seq_ids': {},
            'copy_number': 0,
            'metadata': {},
            'mutations': [],
        },
        'children': [node]
    }


def push_common_mutations_up(tree, first):
    if tree.is_leaf():
        return tree.mutations

    common_muts = None
    for child in tree.children:
        child_muts = push_common_mutations_up(child, first)
        if common_muts is None:
            common_muts = child_muts
        else:
            common_muts = common_muts.intersection(child_muts)

    if len(tree.seq_ids) == 0:
        if first:
            tree.mutations = common_muts or tree.mutations
        else:
            tree.mutations = common_muts.union(tree.mutations)

    return tree.mutations


def remove_parent_mutations(tree):
    for node in tree.traverse(strategy='postorder'):
        if node.up is not None and node.up.name != 'germline':
            node.mutations = node.mutations.difference(node.up.mutations)


def remove_null_nodes(tree):
    removed = False
    for node in tree.traverse():
        if node.up is not None and len(node.mutations) == 0:
            node.up.seq_ids.update(node.seq_ids)
            node.up.copy_number += node.copy_number
            node.delete(prevent_nondicotomic=False)
            removed = True
    return removed


def check_supersets(tree):
    if tree.is_leaf():
        return False

    moved = False
    for c1 in tree.children:
        for c2 in tree.children:
            if c1 == c2:
                continue
            if c1.mutations.issubset(c2.mutations):
                c1.detach()
                c2.add_child(c1)
                moved = True
            elif c2.mutations.issubset(c1.mutations):
                c2.detach()
                c1.add_child(c2)
                moved = True

            overlap = c1.mutations.intersection(c2.mutations)
            if len(overlap) > 0:
                c1.detach()
                c2.detach()
                intermediate = instantiate_node(ete2.Tree(name='NoName'))
                intermediate.mutations = overlap
                intermediate.add_child(c1)
                intermediate.add_child(c2)
                tree.add_child(intermediate)
            moved = moved or check_supersets(c1)

    return moved


def are_null_nodes(tree):
    for node in tree.traverse():
        if node.up is not None and len(node.mutations) == 0:
            return True
    return False


def get_total_muts(tree):
    muts = 0
    for node in tree.traverse():
        muts += len(node.mutations)
    return muts


def cut_tree(tree, max_muts, d=0):
    max_muts -= len(tree.mutations)
    if tree.is_leaf() or max_muts <= 0:
        return [tree]
    subtrees = []
    for node in tree.children:
        subtrees.extend(cut_tree(node, max_muts, d+1))
    return subtrees


def get_seq_pks(tree):
    seq_pks = set([
        (s['sample_id'], s['ai']) for s in tree.seq_ids.values()
    ])
    for child in tree.children:
        seq_pks.update(get_seq_pks(child))
    return seq_pks



from immunedb.common.models import Sequence
from immunedb.util.nested_writer import NestedCSVWriter


class SequenceWriter(object):
    def __init__(self):
        self._required_fields = []

    def set_selected_fields(self, headers):
        self._selected_fields = headers[:]

    def get_required_fields(self, headers):
        for header in self._required_fields:
            if header not in headers:
                headers.append(header)
        return headers

    def preprocess_query(self, sequence_query):
        return sequence_query.order_by(Sequence.ai)

    def format_sequence(self, data, pseudo_seq=False):
        raise NotImplementedError()


class FASTAWriter(SequenceWriter):
    def __init__(self, replaced_sequences):
        self._replaced_sequences = replaced_sequences
        self._required_fields = ['seq_id', 'sequence', 'germline']

    def format_sequence(self, data):
        return '>{}\n{}\n'.format(
            self._get_header(data),
            self._get_sequence(data['sequence'], data['germline'])
        )

    def _get_sequence(self, sequence, germline):
        if not self._replaced_sequences:
            return sequence

        return ''.join(
            [g if s == 'N' else s for s, g in zip(sequence, germline)]
        )

    def _get_header(self, metadata):
        fields = {k: v for k, v in metadata.iteritems() if k in
                  self._selected_fields}
        for field in self._required_fields:
            if field in fields:
                del fields[field]
        return ''.join([
            metadata['seq_id'],
            '|' if len(fields) > 0 else '',
            '|'.join(['{}={}'.format(k, fields[k]) for k in
                     sorted(fields.keys())])
        ])


class FASTQWriter(FASTAWriter):
    def __init__(self, replaced_sequences):
        super(FASTQWriter, self).__init__(replaced_sequences)
        self._required_fields += ['quality']

    def format_sequence(self, data, pseudo_seq=False):
        return '@{}\n{}\n+\n{}\n'.format(
            self._get_header(data),
            self._get_sequence(data['sequence'], data['germline']),
            data['quality'] or ''
        )


class CLIPWriter(FASTAWriter):
    def __init__(self, replaced_sequences):
        super(CLIPWriter, self).__init__(replaced_sequences)
        self._required_fields += ['v_gene', 'j_gene', 'cdr3_aa', 'cdr3_nt',
                                  'cdr3_num_nts', 'clone_germline']
        self._last_cid = None

    def preprocess_query(self, sequence_query):
        return sequence_query.order_by(Sequence.clone_id, Sequence.ai)

    def format_sequence(self, data, pseudo_seq=False):
        if data['clone_id'] != self._last_cid and not pseudo_seq:
            self._last_cid = data['clone_id']
            clone_row = super(CLIPWriter, self).format_sequence({
                'seq_id': 'Germline',
                'v_gene': data['v_gene'],
                'j_gene': data['j_gene'],
                'cdr3_aa': data['cdr3_aa'],
                'cdr3_nt': data['cdr3_nt'],
                'cdr3_len': data['cdr3_num_nts'],
                'sequence': data['sequence'],
                'germline': data['clone_germline']
            })
        else:
            clone_row = ''
        return clone_row + super(CLIPWriter, self).format_sequence(data)


class CSVWriter(SequenceWriter):
    def set_selected_fields(self, headers):
        super(CSVWriter, self).set_selected_fields(headers)
        self._csv = NestedCSVWriter(headers, streaming=True)

    def format_sequence(self, data, pseudo_seq=False):
        return self._csv.add_row(data)



import os

import multiprocessing as mp
import traceback
import Queue

import logging
from immunedb.util.log import logger


class Worker(object):
    def log(self, lvl, msg):
        logger.log(lvl, 'Worker {}: {}'.format(self._worker_id, msg))

    def info(self, msg):
        self.log(logging.INFO, msg)

    def error(self, msg):
        self.log(logging.ERROR, msg)

    def warning(self, msg):
        self.log(logging.WARNING, msg)

    def do_task(self, args):
        raise NotImplementedError

    def cleanup(self):
        pass


class TaskQueue(object):
    def __init__(self):
        self._task_queue = mp.JoinableQueue()
        self._num_tasks = 0
        self._workers = []

    def add_task(self, args):
        self._num_tasks += 1
        self._task_queue.put(args)

    def add_tasks(self, tasks):
        for i, task in enumerate(tasks):
            self.add_task(task)

    def add_worker(self, worker):
        self._workers.append(
            mp.Process(
                target=self._func_wrap,
                args=(len(self._workers) + 1, worker)
            )
        )

    def add_workers(self, count, worker_cls, *args, **kwargs):
        for _ in range(count):
            self.add_worker(worker_cls(*args, **kwargs))

    def join(self):
        self._task_queue.join()

    def signal_end(self):
        for _ in self._workers:
            self.add_task(None)

    def start(self, block=True):
        if block:
            self.signal_end()

        for worker in self._workers:
            worker.start()
        if block:
            self.join()

    def _func_wrap(self, worker_id, worker):
        worker._worker_id = worker_id
        while True:
            try:
                args = self._task_queue.get()
                if args is None:
                    self._task_queue.task_done()
                    break
                else:
                    worker.do_task(args)
                    self._task_queue.task_done()
            except Exception:
                worker.error(
                    'The task was not completed because:\n{}'.format(
                        traceback.format_exc()))
                self._task_queue.task_done()
        worker.cleanup()

    def num_tasks(self):
        return self._num_tasks


# V2 of multiprocessing
class _EndOfDataSentinel(object):
    pass


class EndOfDataException(Exception):
    pass


class SizedQueue(object):
    def __init__(self):
        self.queue = mp.Queue()

    def put(self, val):
        self.queue.put(val)

    def get(self):
        res = self.queue.get()
        if res == _EndOfDataSentinel:
            raise EndOfDataException()
        return res

    def __iter__(self):
        while True:
            try:
                yield self.get()
            except EndOfDataException:
                break


def _process_wrapper(process_func, in_queue, out_queue, **kwargs):
    logger.info('Starting worker with PID {}'.format(os.getpid()))
    while True:
        try:
            for d in process_func(in_queue.get(), **kwargs):
                out_queue.put(d)
        except Queue.Empty:
            continue
        except EndOfDataException:
            break
        except Exception:
            logging.warning('Error in process_func:\n{}'.format(
                traceback.format_exc()))
    logger.info('Stopping worker with PID {}'.format(os.getpid()))


def _agg_wrapper(aggregate_func, aggregate_queue, return_val, **kwargs):
    ret = aggregate_func(aggregate_queue, **kwargs)
    if ret:
        return_val.update(ret)


def process_data(generate_func_or_iter,
                 process_func,
                 aggregate_func,
                 nproc,
                 generate_args={},
                 process_args={},
                 aggregate_args={}):
    process_queue = SizedQueue()
    aggregate_queue = SizedQueue()
    return_val = mp.Manager().dict()

    workers = []
    for i in range(nproc):
        w = mp.Process(
            target=_process_wrapper,
            args=(process_func, process_queue, aggregate_queue),
            kwargs=process_args
        )
        w.start()
        workers.append(w)

    agg_p = mp.Process(
        target=_agg_wrapper,
        args=(aggregate_func, aggregate_queue, return_val,),
        kwargs=aggregate_args
    )
    agg_p.start()

    try:
        input_p = None
        iter(generate_func_or_iter)
        for v in generate_func_or_iter:
            process_queue.put(v)
    except TypeError:
        input_p = mp.Process(
            target=generate_func_or_iter,
            args=(process_queue,),
            kwargs=generate_args
        )
        input_p.start()

    if input_p:
        input_p.join()

    for n in range(nproc):
        process_queue.put(_EndOfDataSentinel)

    for i, w in enumerate(workers):
        w.join()
        logger.info('{} workers finished'.format(i + 1))
    aggregate_queue.put(_EndOfDataSentinel)

    logger.info('Waiting on aggregation')
    agg_p.join()
    logger.info('Done aggregation')

    return return_val.copy()



from collections import OrderedDict

import dnautils
import re

from Bio import SeqIO
from Bio.Seq import Seq

from immunedb.common.models import CDR3_OFFSET
from immunedb.util.hyper import hypergeom
from immunedb.identification import AlignmentException, get_common_seq


class GermlineException(Exception):
    pass


class GeneName(object):
    def __init__(self, name):
        self.name = name
        parts = re.search('((([A-Z]+)(\d+)([^*]+)?)(\*(\d+))?)',
                          self.name).groups()
        self.name = parts[0]
        self.base = parts[1]
        self.prefix = parts[2]
        self.family = parts[3]
        self.allele = parts[6] if parts[6] else None

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __cmp__(self, other):
        return cmp(self.name, other.name)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        return ('<GeneName={}, base={}, prefix={}, family={}, '
                'allele={}>').format(str(self), self.base, self.prefix,
                                     self.family, self.allele)


class GeneTies(dict):
    TIES_PROB_THRESHOLD = 0.01

    def __init__(self, genes, remove_gaps=True, no_ties=False):
        self.ties = {}
        self.hypers = {}
        self.remove_gaps = remove_gaps
        self.no_ties = no_ties

        self.update(genes)

        self.allele_lookup = {}
        for name in self.keys():
            self.allele_lookup[name] = set([])
            for name2 in self.keys():
                if name2.base == name.base:
                    self.allele_lookup[name].add(name2)

    def all_ties(self, length, mutation, cutoff=True):
        ties = {}
        for name in self:
            tie_name = tuple(sorted(self.get_ties([name], length, mutation)))
            if tie_name not in ties:
                ties[tie_name] = get_common_seq(
                    [self[n] for n in tie_name], cutoff=cutoff
                )
        return ties

    def get_ties(self, genes, length, mutation):
        ties = set([])
        for gene in genes:
            ties.update(self.get_single_tie(gene, length, mutation))
        return ties

    def get_single_tie(self, gene, length, mutation):
        # Used to disable gene ties for genotyping
        if self.no_ties:
            return set([gene])
        length = int(length)
        mutation = round(mutation, 3)
        mutation = self.mut_bucket(mutation)
        key = (length, mutation)

        if key not in self.ties:
            self.ties[key] = {}

        if gene not in self:
            return set([gene])

        if gene not in self.ties[key]:
            s_1 = (
                self[gene].replace('-', '') if self.remove_gaps else self[gene]
            )
            self.ties[key][gene] = set([gene])

            for name, v in sorted(self.iteritems()):
                s_2 = v.replace('-', '') if self.remove_gaps else v
                K = dnautils.hamming(s_1[-length:], s_2[-length:])
                p = self._hypergeom(length, mutation, K)
                if p >= self.TIES_PROB_THRESHOLD:
                    self.ties[key][gene].add(name)
            self.ties[key][gene] = self.all_alleles(self.ties[key][gene])

        return self.ties[key][gene]

    def _hypergeom(self, length, mutation, K):
        key = (length, mutation, K)
        if key not in self.hypers:
            self.hypers[key] = hypergeom(length, mutation, K)
        return self.hypers[key]

    def mut_bucket(self, mut):
        if 0 <= mut <= .05:
            return .05
        if mut <= .15:
            return .15
        return .30

    def all_alleles(self, genes):
        all_genes = set([])
        for gene in genes:
            all_genes.update(self.allele_lookup[gene])
        return all_genes


class VGermlines(GeneTies):
    def __init__(self, path_to_germlines, **kwargs):
        self._min_length = None
        self.alignments = OrderedDict()

        with open(path_to_germlines) as fh:
            for record in SeqIO.parse(fh, 'fasta'):
                if record.seq.startswith('-'):
                    continue
                try:
                    v = VGene(str(record.seq))
                    name = GeneName(record.id)
                    self.alignments[name] = v
                    self[name] = str(record.seq)
                    if (self._min_length is None or
                            self._min_length > len(v.sequence_ungapped)):
                        self._min_length = len(v.sequence_ungapped)
                except Exception:
                    continue

        super(VGermlines, self).__init__({k: v for k, v in self.iteritems()},
                                         **kwargs)

    def get_single_tie(self, gene, length, mutation):
        return super(VGermlines, self).get_single_tie(
            gene, min(self.length_bucket(length), self._min_length), mutation
        )

    def length_bucket(self, length):
        if 0 < length <= 100:
            return 100
        if 100 < length <= 150:
            return 150
        if 150 < length <= 200:
            return 200
        return 300


class VGene(object):
    def __init__(self, gapped_sequence):
        self.sequence = str(gapped_sequence).upper()
        self.sequence_ungapped = self.sequence.replace('-', '')
        if self.sequence[CDR3_OFFSET:].count('-') > 0:
            raise AlignmentException('Cannot have gaps after CDR3 start '
                                     '(position {})'.format(CDR3_OFFSET))
        try:
            self.ungapped_anchor_pos = find_v_position(
                self.sequence_ungapped).next()
        except StopIteration:
            raise AlignmentException('Unable to find anchor')

    def align(self, other_v):
        diff = abs(self.ungapped_anchor_pos - other_v.ungapped_anchor_pos)
        this_seq = self.sequence_ungapped
        other_seq = other_v.sequence_ungapped

        # Trim the sequence which has the maximal anchor position, and
        # determine the CDR3 start position without gaps
        if self.ungapped_anchor_pos > other_v.ungapped_anchor_pos:
            this_seq = this_seq[diff:]
            cdr3_start = self.ungapped_anchor_pos - diff
        else:
            other_seq = other_seq[diff:]
            cdr3_start = other_v.ungapped_anchor_pos - diff

        return {
            'base': this_seq,
            'seq': other_seq,
            'diff': diff,
            'cdr3_start': cdr3_start
        }

    def compare(self, other_v, max_extent, max_streak):
        alignment = self.align(other_v)
        this_seq = alignment['base'][:max_extent]
        other_seq = alignment['seq'][:max_extent]
        cdr3_offset = alignment['cdr3_start']

        # Determine the CDR3 in the germline and sequence
        this_cdr3 = this_seq[cdr3_offset:]
        other_cdr3 = other_seq[cdr3_offset:]
        length = min(len(this_cdr3), len(other_cdr3))
        this_cdr3 = this_cdr3[:length]
        other_cdr3 = other_cdr3[:length]
        if len(this_cdr3) == 0 or len(other_cdr3) == 0:
            raise AlignmentException('Empty CDR3 found after alignment')

        # Find the extent of the sequence's V into the CDR3
        streak = dnautils.find_streak_position(
            this_cdr3, other_cdr3, max_streak)
        if streak is not None:
            # If there is a streak of mismatches, cut after the streak
            max_index = cdr3_offset + (streak - max_streak)
        else:
            # Unlikely: the CDR3 in the sequence exactly matches the
            # germline.  Use the smaller sequence length (full match)
            max_index = cdr3_offset + min(len(this_cdr3), len(other_cdr3))
        # Compare to the end of V
        this_seq = this_seq[:max_index]
        other_seq = other_seq[:max_index]

        if len(this_seq) != len(other_seq) or len(this_seq) == 0:
            raise AlignmentException('Unequal sequences after alignment')
        # Determine the distance between the germline and sequence
        dist = dnautils.hamming(this_seq, other_seq)

        return dist, len(other_seq)


def find_v_position(sequence):
    if type(sequence) == str:
        sequence = Seq(sequence)
    frames = []
    for shift in [2, 1, 0]:
        seq = sequence[shift:]
        seq = seq[:len(seq) - len(seq) % 3]
        frames.append((shift, str(seq.translate())))

    # Try to find DxxxyzC
    for found in _find_with_frameshifts(frames, 'D(.{3}((YY)|(YC)|(YH)))C'):
        yield found
    # Try to find 'YYC', 'YCC', or 'YHC'
    for found in _find_with_frameshifts(frames, 'Y([YHC])C'):
        yield found
    # Try to find 'DxxxxxC'
    for found in _find_with_frameshifts(frames, 'D(.{5})C'):
        yield found


def _find_with_frameshifts(frames, regex):
    for (shift, aas) in frames:
        res = re.search(regex, aas)
        if res is not None:
            yield (res.end() - 1) * 3 + shift


class JGermlines(GeneTies):
    defaults = {
        'upstream_of_cdr3': 31,
        'anchor_len': 18,
        'min_anchor_len': 12,
    }

    def __init__(self, path_to_germlines,
                 upstream_of_cdr3=defaults['upstream_of_cdr3'],
                 anchor_len=defaults['anchor_len'],
                 min_anchor_len=defaults['min_anchor_len'],
                 **kwargs):
        self._upstream_of_cdr3 = upstream_of_cdr3
        self._anchor_len = anchor_len
        self._min_anchor_len = min_anchor_len
        self._min_length = None

        with open(path_to_germlines) as fh:
            for record in SeqIO.parse(fh, 'fasta'):
                name = GeneName(record.id)
                if all(map(lambda c: c in 'ATCGN', record.seq.upper())):
                    self[name] = str(record.seq).upper()
                    if (self._min_length is None or
                            len(self[name]) < self._min_length):
                        self._min_length = len(self[name])

        self._anchors = {name: seq[-anchor_len:] for name, seq in
                         self.iteritems()}
        super(JGermlines, self).__init__({k: v for k, v in self.iteritems()},
                                         **kwargs)

    @property
    def upstream_of_cdr3(self):
        return self._upstream_of_cdr3

    @property
    def anchor_len(self):
        return self._anchor_len

    @property
    def full_anchors(self):
        return self._anchors

    def get_j_in_cdr3(self, gene):
        return self[gene][:-self._upstream_of_cdr3]

    def get_all_anchors(self, allowed_genes=None):
        if allowed_genes is None:
            allowed_genes = self
        else:
            allowed_genes = {k: v for k, v in self.iteritems() if k.name in
                             allowed_genes}
        max_len = max(map(len, allowed_genes.values()))
        for trim_len in range(0, max_len, 3):
            for j, seq in allowed_genes.iteritems():
                trimmed_seq = seq[-self.anchor_len:-trim_len]
                if len(trimmed_seq) >= self._min_anchor_len:
                    yield trimmed_seq, j

    def get_single_tie(self, gene, length, mutation):
        # Used to disable gene ties for genotyping
        if self.no_ties:
            return set([gene])
        seq = self[gene][-self.anchor_len:]
        tied = self.all_alleles(set([gene]))
        for j, other_seq in sorted(self.iteritems()):
            other_seq = other_seq[-self.anchor_len:][:len(seq)]
            if other_seq == seq:
                tied.add(j)
            elif dnautils.hamming(other_seq, seq) == 0:
                tied.add(j)
        return tied

    def all_ties(self, length, mutation):
        ties = {}
        for name in self:
            tie_name = tuple(sorted(self.get_ties([name], length, mutation)))
            if tie_name not in ties:
                ties[tie_name] = get_common_seq(
                    [self[n] for n in tie_name], right=True
                )
        return ties



import os
import sys
import time
import pygtrie
from sqlalchemy import func

from Bio import SeqIO

import dnautils

import immunedb.common.config as config
import immunedb.common.modification_log as mod_log
from immunedb.common.models import (Sample, SampleMetadata, Sequence, NoResult,
                                    Study, Subject)
from immunedb.identification import (add_noresults_for_vdj, add_sequences,
                                     AlignmentException)
from immunedb.identification.anchor import AnchorAligner
from immunedb.identification.metadata import (MetadataException,
                                              parse_metadata, REQUIRED_FIELDS)
from immunedb.identification.vdj_sequence import VDJSequence
from immunedb.identification.genes import JGermlines, VGermlines
import immunedb.util.concurrent as concurrent
import immunedb.util.funcs as funcs
from immunedb.util.log import logger


class IdentificationProps(object):
    defaults = {
        'max_v_ties': 50,
        'min_similarity': .60,
        'max_padding': None,
        'trim_to': 0,
        'allow_cross_family': False,
        'max_insertions': 5,
        'max_deletions': 5,
        'genotyping': False
    }

    def __init__(self, **kwargs):
        for prop, default in self.defaults.iteritems():
            setattr(self, prop, kwargs.get(prop, default))

    def valid_v_ties(self, alignment):
        return len(alignment.v_gene) <= self.max_v_ties

    def valid_min_similarity(self, alignment):
        return (alignment.v_match / float(alignment.v_length) >=
                self.min_similarity)

    def valid_padding(self, alignment):
        return (self.max_padding is None or
                alignment.seq_start <= self.max_padding)

    def valid_families(self, alignment):
        if self.allow_cross_family:
            return True

        family = None
        for gene in alignment.v_gene:
            if not family:
                family = gene.family
            elif gene.family != family:
                return False
        return True

    def valid_indels(self, alignment):
        return (len(alignment.insertions) <= self.max_insertions and
                len(alignment.deletions) <= self.max_deletions)

    def validate_cdr3(self, alignment):
        return alignment.cdr3_num_nts >= 9

    def validate(self, alignment):
        if not self.valid_min_similarity(alignment):
            raise AlignmentException(
                'V-identity too low {} < {}'.format(
                    alignment.v_match / float(alignment.v_length),
                    self.min_similarity))
        if not self.valid_v_ties(alignment):
            raise AlignmentException('Too many V-ties {} > {}'.format(
                len(alignment.v_gene), self.max_v_ties))
        if not self.valid_padding(alignment):
            raise AlignmentException('Too much padding {} (max {})'.format(
                alignment.seq_start, self.max_padding))
        if not self.valid_families(alignment):
            raise AlignmentException('Cross-family V-call')
        if not self.valid_indels(alignment):
            raise AlignmentException(
                'Too many indels insertions={} deletions={}'.format(
                    alignment.insertions, alignment.deletions))
        if not self.validate_cdr3(alignment):
            raise AlignmentException('CDR3 too short {}'.format(
                alignment.cdr3_num_nts))


def setup_sample(session, meta):
    study, new = funcs.get_or_create(session, Study, name=meta['study_name'])

    if new:
        logger.info('Created new study "{}"'.format(study.name))
        session.commit()

    name = meta['sample_name']
    sample, new = funcs.get_or_create(session, Sample, name=name, study=study)

    if new:
        subject, new = funcs.get_or_create(
            session, Subject, study=study,
            identifier=meta['subject'])
        sample.subject = subject

        for key, value in meta.iteritems():
            if key not in REQUIRED_FIELDS:
                session.add(SampleMetadata(
                    sample=sample, key=key, value=value
                ))

    session.commit()
    return sample


def process_vdj(vdj, aligner):
    try:
        alignment = aligner.get_alignment(vdj)
        return [{
            'status': 'success',
            'alignment': alignment
        }]
    except AlignmentException as e:
        return [{
            'status': 'noresult',
            'vdj': vdj,
            'reason': str(e)
        }]
    except Exception as e:
        return [{
            'status': 'error',
            'vdj': vdj,
            'reason': str(e)
        }]


def aggregate_vdj(aggregate_queue):
    alignments = {
        'success': {},
        'noresult': []
    }
    for result in aggregate_queue:
        if result['status'] == 'success':
            alignment = result['alignment']
            seq_key = alignment.sequence.sequence
            if seq_key in alignments['success']:
                alignments['success'][seq_key].sequence.ids.extend(
                    alignment.sequence.ids)
            else:
                alignments['success'][seq_key] = alignment
        elif result['status'] == 'noresult':
            alignments['noresult'].append(result)
        elif result['status'] == 'error':
            logger.error(
                'Unexpected error processing sequence {}\n\t{}'.format(
                    result['vdj'].ids[0], result['reason']))
    alignments['success'] = alignments['success'].values()
    return alignments


def process_vties(alignment, aligner, avg_len, avg_mut, props):
    try:
        aligner.align_to_germline(alignment, avg_len, avg_mut)
        if props.trim_to:
            alignment.trim_to(props.trim_to)

        props.validate(alignment)
        return [{
            'status': 'success',
            'alignment': alignment
        }]
    except AlignmentException as e:
        return [{
            'status': 'noresult',
            'alignment': alignment,
            'reason': str(e)
        }]
    except Exception as e:
        return [{
            'status': 'error',
            'alignment': alignment,
            'reason': str(e)
        }]


def aggregate_vties(aggregate_queue):
    bucketed_seqs = {
        'success': {},
        'noresult': []
    }
    for result in aggregate_queue:
        if result['status'] == 'success':
            alignment = result['alignment']
            bucket_key = (
                funcs.format_ties(alignment.v_gene),
                funcs.format_ties(alignment.j_gene),
                len(alignment.cdr3)
            )

            bucket = bucketed_seqs['success'].setdefault(bucket_key, {})
            if alignment.sequence.sequence in bucket:
                bucket[alignment.sequence.sequence].sequence.ids += (
                    alignment.sequence.ids
                )
            else:
                bucket[alignment.sequence.sequence] = alignment
        elif result['status'] == 'noresult':
            bucketed_seqs['noresult'].append(result)
        elif result['status'] == 'error':
            logger.error(
                'Unexpected error processing sequence {}\n\t{}'.format(
                    result['alignment'].sequence.ids[0]))

    bucketed_seqs['success'] = [
        b.values() for b in bucketed_seqs['success'].values()
    ]
    return bucketed_seqs


def process_collapse(sequences):
    sequences = sorted(
        sequences,
        key=lambda s: (len(s.sequence.ids), s.sequence.ids[0]),
        reverse=True
    )
    uniques = []
    while len(sequences) > 0:
        larger = sequences.pop(0)
        for i in reversed(range(len(sequences))):
            smaller = sequences[i]
            if dnautils.equal(larger.sequence.sequence,
                              smaller.sequence.sequence):
                larger.sequence.ids += smaller.sequence.ids
                del sequences[i]
        uniques.append(larger)
    return uniques


def aggregate_collapse(aggregate_queue, db_config, sample_id, props):
    bulk_add = []
    session = config.init_db(db_config, create=False)
    sample = session.query(Sample).filter(Sample.id == sample_id).one()
    for i, alignment in enumerate(aggregate_queue):
        bulk_add.append(alignment)
        if len(bulk_add) >= 1000:
            add_sequences(session, bulk_add, sample,
                          strip_alleles=not props.genotyping)
            bulk_add = []
            logger.info('Processed {}'.format(i + 1))
            session.commit()
    if bulk_add:
        add_sequences(session, bulk_add, sample,
                      strip_alleles=not props.genotyping)
    logger.info('Finished aggregating sequences')
    session.commit()
    session.close()


def collapse_exact(process_queue, path):
    vdjs = pygtrie.StringTrie()
    parser = SeqIO.parse(path, 'fasta' if path.endswith('.fasta') else 'fastq')

    # Collapse identical sequences
    logger.info('Collapsing identical sequences')
    for record in parser:
        try:
            seq = str(record.seq)
            if seq not in vdjs:
                vdjs[seq] = VDJSequence(
                    ids=[],
                    sequence=seq,
                    quality=funcs.ord_to_quality(
                        record.letter_annotations.get('phred_quality')
                    )
                )
            vdjs[seq].ids.append(record.description)
        except ValueError:
            continue

    logger.info('There are {} (unaligned) unique sequences'.format(len(vdjs)))
    for v in vdjs.values():
        process_queue.put(v)


def process_sample(db_config, v_germlines, j_germlines, path, meta, props,
                   nproc):
    session = config.init_db(db_config)
    start = time.time()
    logger.info('Starting sample {}'.format(meta['sample_name']))
    sample = setup_sample(session, meta)

    aligner = AnchorAligner(v_germlines, j_germlines)

    # Initial VJ assignment
    alignments = concurrent.process_data(
        collapse_exact,
        process_vdj,
        aggregate_vdj,
        nproc,
        process_args={'aligner': aligner},
        generate_args={'path': path},
    )
    logger.info('Adding noresults')
    for result in alignments['noresult']:
        add_noresults_for_vdj(session, result['vdj'], sample, result['reason'])

    alignments = alignments['success']
    if alignments:
        avg_len = (
            sum([v.v_length for v in alignments]) /
            float(len(alignments)))
        avg_mut = (
            sum([v.v_mutation_fraction for v in alignments]) /
            float(len(alignments))
        )
        sample.v_ties_mutations = avg_mut
        sample.v_ties_len = avg_len
        logger.info('Re-aligning {} sequences to V-ties: Mutations={}, '
                    'Length={}'.format(len(alignments),
                                       round(avg_mut, 2),
                                       round(avg_len, 2)))
        session.commit()
        # Realign to V-ties
        v_ties = concurrent.process_data(
            alignments,
            process_vties,
            aggregate_vties,
            nproc,
            process_args={'aligner': aligner, 'avg_len': avg_len, 'avg_mut':
                          avg_mut, 'props': props},
        )
        logger.info('Adding noresults')

        for result in funcs.periodic_commit(session, v_ties['noresult'], 100):
            add_noresults_for_vdj(session, result['alignment'].sequence,
                                  sample, result['reason'])

        logger.info('Collapsing {} buckets'.format(len(v_ties['success'])))
        session.commit()
        concurrent.process_data(
            v_ties['success'],
            process_collapse,
            aggregate_collapse,
            nproc,
            aggregate_args={'db_config': db_config, 'sample_id': sample.id,
                            'props': props}
        )
        session.expire_all()
        session.commit()

        identified = int(session.query(
            func.sum(Sequence.copy_number)
        ).filter(
            Sequence.sample == sample
        ).scalar() or 0)
        noresults = int(session.query(
            func.count(NoResult.pk)
        ).filter(
            NoResult.sample == sample
        ).scalar() or 0)
        logger.info(
            'Completed sample {} in {}m - {}/{} ({}%) identified'.format(
                sample.name,
                round((time.time() - start) / 60., 1),
                identified,
                identified + noresults,
                int(100 * identified / float(identified + noresults))
            )
        )
    session.close()


def run_identify(session, args):
    mod_log.make_mod('identification', session=session, commit=True,
                     info=vars(args))
    # Load the germlines from files
    v_germlines = VGermlines(args.v_germlines, no_ties=args.genotyping)
    j_germlines = JGermlines(args.j_germlines, args.upstream_of_cdr3,
                             args.anchor_len, args.min_anchor_len,
                             no_ties=args.genotyping)

    # If metadata is not specified, assume it is "metadata." in the
    # directory
    meta_fn = args.metadata if args.metadata else os.path.join(
        args.sample_dir, 'metadata.tsv')

    # Verify the metadata file exists
    if not os.path.isfile(meta_fn):
        logger.error('Metadata file not found.')
        sys.exit(-1)

    with open(meta_fn, 'rU') as fh:
        try:
            metadata = parse_metadata(session, fh, args.warn_existing,
                                      args.warn_missing, args.sample_dir)
        except MetadataException as ex:
            logger.error(ex.message)
            sys.exit(-1)

    session.close()
    # Create the tasks for each file
    props = IdentificationProps(**args.__dict__)
    for sample_name in sorted(metadata.keys()):
        process_sample(
            args.db_config, v_germlines, j_germlines,
            os.path.join(
                args.sample_dir,
                metadata[sample_name]['file_name']
            ),
            metadata[sample_name],
            props,
            args.nproc
        )



import json

import immunedb.common.config as config
from immunedb.common.models import Clone, Sequence, SequenceCollapse
import immunedb.common.modification_log as mod_log
from immunedb.trees import PhylogeneticTree, tree_as_dict
import immunedb.util.concurrent as concurrent
from immunedb.util.log import logger


class ClearcutWorker(concurrent.Worker):
    def __init__(self, session, tree_prog, min_count, min_samples,
                 min_seq_copies, exclude_stops):
        self._session = session
        self._tree_prog = tree_prog
        self._min_count = min_count
        self._min_samples = min_samples
        self._min_seq_copies = min_seq_copies
        self._exclude_stops = exclude_stops

    def do_task(self, clone_id):
        clone_inst = self._session.query(Clone).filter(
            Clone.id == clone_id).first()
        if clone_inst is None:
            return

        self.info('Running clone {}'.format(clone_inst.id))

        sequences = self._session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.clone_id == clone_id,
            SequenceCollapse.copy_number_in_subject > self._min_seq_copies
        )

        if self._exclude_stops:
            sequences = sequences.filter(Sequence.stop == 0)

        sequences = sequences.order_by(Sequence.v_length)

        try:
            tree = PhylogeneticTree(clone_inst.consensus_germline, sequences)
            tree.run(self._session, self._tree_prog)
        except Exception as e:
            logger.error('Error running clone {}: {}'.format(clone_id, e))
            return

        final = {
            'info': {
                'min_count': self._min_count,
                'min_samples': self._min_samples,
                'min_seq_copies': self._min_seq_copies,
                'exclude_stops': self._exclude_stops
            },
            'tree': tree_as_dict(tree.tree)
        }
        clone_inst.tree = json.dumps(final)
        self._session.add(clone_inst)
        self._session.commit()


def run_clearcut(session, args):
    if args.clone_ids is not None:
        clones = session.query(Clone.id).filter(
            Clone.id.in_(args.clone_ids))
    else:
        if args.subject_ids is not None:
            clones = session.query(Clone.id).filter(
                Clone.subject_id.in_(args.subject_ids))
        else:
            clones = session.query(Clone.id)

    if not args.force:
        clones = clones.filter(Clone.tree.is_(None))
    clones = [c.id for c in clones]
    mod_log.make_mod('clone_tree', session=session, commit=True,
                     info=vars(args))

    tasks = concurrent.TaskQueue()

    logger.info('Creating task queue for clones')
    for clone_id in clones:
        tasks.add_task(clone_id)

    for _ in range(0, args.nproc):
        session = config.init_db(args.db_config)
        tasks.add_worker(ClearcutWorker(session, args.clearcut_path,
                                        args.min_count, args.min_samples,
                                        args.min_seq_copies,
                                        args.exclude_stops))

    tasks.start()



from collections import Counter, OrderedDict
import csv
import re
from sqlalchemy.orm import joinedload

from immunedb.identification.genes import GeneName
from immunedb.common.models import (Clone, CloneStats, Sample, Sequence,
                                    SequenceCollapse, Subject)
import immunedb.util.funcs as funcs
from immunedb.util.log import logger
from immunedb.util.lookups import aas_from_nts


def export_vdjtools(session, args):
    fieldnames = ['count', 'freq', 'cdr3nt', 'cdr3aa', 'v', 'd', 'j']
    if args.include_uniques:
        fieldnames.append('unique')

    clone_features = {c.id: (c.v_gene, c.j_gene, c.cdr3_nt)
                      for c in session.query(Clone.id, Clone.v_gene,
                                             Clone.j_gene, Clone.cdr3_nt)}
    for sample in session.query(Sample).order_by(Sample.id):
        logger.info('Exporting sample {}'.format(sample.name))
        sample_clones = {}
        stats = session.query(
            CloneStats.clone_id, CloneStats.total_cnt, CloneStats.unique_cnt
        ).filter(
            CloneStats.sample_id == sample.id
        )
        for stat in stats:
            key = clone_features[stat.clone_id]
            sample_clones.setdefault(key, Counter())['total'] += stat.total_cnt
            sample_clones[key]['unique'] += stat.unique_cnt

        writer = csv.DictWriter(
            open('{}.sample.txt'.format(sample.name), 'w+'),
            fieldnames=fieldnames,
            delimiter='\t',
            extrasaction='ignore'
        )
        total = float(sum([c['total'] for c in sample_clones.values()]))
        writer.writeheader()
        for key in sorted(sample_clones, key=sample_clones.get, reverse=True):
            counts = sample_clones[key]
            if counts['total'] < args.min_clone_size:
                continue
            v, j, cdr3_nt = key
            writer.writerow({
                'count': counts['total'],
                'freq': counts['total'] / total,
                'cdr3nt': cdr3_nt,
                'cdr3aa': aas_from_nts(cdr3_nt),
                'v': v,
                'd': '.',
                'j': j,
                'unique': counts['unique']
            })


def get_genbank_entries(seq, inference, gene_db):
    entry = OrderedDict()
    trim_start = re.search('[N-]*', seq.sequence).end()
    trim_end = re.search('[N-]*', seq.sequence[::-1]).end()
    trimmed_seq = seq.sequence[trim_start:len(seq.sequence) - trim_end]
    # Offsets of regions/segments
    v_gaps = trimmed_seq.count('-')
    trimmed_seq = trimmed_seq.replace('-', '')
    v_region = (1, len(trimmed_seq), 'V_region')
    v_segment = (1, 309 - v_gaps - trim_start, 'V_segment')
    cds = (
        v_segment[1] + 1,
        v_segment[1] + seq.cdr3_num_nts,
        'CDS'
    )
    j_segment = (cds[1] + 1, v_region[1], 'J_segment')

    # Key
    entry[('>Feature', seq.seq_id.split(' ')[0])] = []
    # V region
    entry[v_region] = []
    # V segment
    first_v = seq.v_gene.split('|')[0]
    entry[v_segment] = (
        ('gene', first_v),
        ('db_xref', '{}:{}'.format(gene_db, first_v)),
        ('inference', inference),
    )

    # J segment
    first_j = seq.j_gene.split('|')[0]
    entry[j_segment] = (
        ('gene', first_j),
        ('db_xref', '{}:{}'.format(gene_db, first_j)),
        ('inference', inference),
    )

    # CDS
    cds = ('<{}'.format(cds[0]), '>{}'.format(cds[1]), 'CDS')
    entry[cds] = (
        ('function', 'junction'),
        ('codon_start', 1),
        ('product', 'immunoglobulin heavy chain variable domain'),
        ('inference', inference)
    )

    return entry, trimmed_seq


def export_sample_genbank(session, sample_id, gene_db, inference, header):
    for sample in session.query(Sample):
        seqs = session.query(
            Sequence.seq_id,
            Sequence.v_gene,
            Sequence.j_gene,
            Sequence.cdr3_num_nts,
            Sequence.sequence,
            Sequence.copy_number
        ).filter(
            Sequence.sample_id == sample_id
        ).filter(Sequence.stop == 0)
        with open('{}.tbl'.format(sample.name), 'w+') as gb_fh:
            writer = csv.writer(gb_fh, delimiter='\t')
            with open('{}.fsa'.format(sample.name), 'w+') as fasta_fh:
                for seq in seqs:
                    gb_entry, fasta_seq = get_genbank_entries(
                        seq, inference, gene_db)
                    for entry, indented in gb_entry.iteritems():
                        writer.writerow(entry)
                        if indented:
                            for indent in indented:
                                writer.writerow(('', '', '') + indent)
                    seq_header = header + ' [note=AIRR_READ_COUNT:{}]'.format(
                        seq.copy_number)
                    fasta_fh.write('>{}\n{}\n'.format(
                        seq.seq_id.split(' ')[0] + ' {}'.format(seq_header),
                        fasta_seq))


def export_changeo(session, args):
    for subject in session.query(Subject):
        logger.info('Exporting subject {}'.format(subject.identifier))
        seqs = session.query(Sequence).filter(
            Sequence.subject_id == subject.id
        ).join(
            SequenceCollapse
        ).options(
            joinedload(Sequence.clone),
            joinedload(Sequence.collapse)
        )
        if args.clones_only:
            seqs = seqs.filter(~Sequence.clone_id.is_(None))
        if args.min_subject_copies is not None:
            seqs = seqs.filter(
                SequenceCollapse.copy_number_in_subject >=
                args.min_subject_copies
            )

        with open('{}.changeo.tsv'.format(subject.identifier), 'w+') as fh:
            writer = csv.DictWriter(
                fh, delimiter='\t', extrasaction='ignore', fieldnames=[
                    'SEQUENCE_ID',
                    'SEQUENCE_IMGT',
                    'FUNCTIONAL',
                    'IN_FRAME',
                    'STOP',
                    'V_CALL',
                    'J_CALL',
                    'JUNCTION_LENGTH',
                    'JUNCTION',
                    'V_SCORE',
                    'V_IDENTITY',
                    'J_SCORE',
                    'J_IDENTITY',
                    'DUPCOUNT',
                    'CLONE',
                    'CLONE_CDR3_NT'
                ]
            )
            writer.writeheader()
            for seq in funcs.yield_limit(seqs, Sequence.ai):
                v_prefix = GeneName(seq.v_gene).prefix
                j_prefix = GeneName(seq.j_gene).prefix
                writer.writerow({
                    'SEQUENCE_ID': seq.seq_id,
                    'SEQUENCE_IMGT': seq.sequence,
                    'FUNCTIONAL': 'T' if seq.functional else 'F',
                    'IN_FRAME': 'T' if seq.in_frame else 'F',
                    'STOP': 'T' if seq.stop else 'F',
                    'V_CALL': seq.v_gene.replace('|', ',' + v_prefix),
                    'J_CALL': seq.j_gene.replace('|', ',' + j_prefix),
                    'JUNCTION_LENGTH': seq.cdr3_num_nts,
                    'JUNCTION': seq.cdr3_nt,
                    'V_SCORE': round(
                        100 * seq.v_match / float(seq.v_length), 2),
                    'V_IDENTITY': seq.v_match,
                    'J_SCORE': round(
                        100 * seq.j_match / float(seq.j_length), 2),
                    'J_IDENTITY': seq.j_match,
                    'DUPCOUNT': seq.collapse.copy_number_in_subject,
                    'CLONE': seq.clone_id,
                    'CLONE_CDR3_NT': seq.clone.cdr3_nt if seq.clone else ''
                })


def export_genbank(session, args):
    args.inference = 'alignment:' + args.inference

    header = (
        '[organism={}] '
        '[moltype={}] '
        '[keywords=AIRR]').format(
        args.species, args.mol_type)
    samples = args.sample_ids or [s.id for s in session.query(Sample.id).all()]

    for sample_id in samples:
        logger.info('Exporting sample {}'.format(sample_id))
        export_sample_genbank(session, sample_id, args.gene_db, args.inference,
                              header)



import csv
import os
import re
import subprocess
import shlex
import StringIO

from sqlalchemy import desc, text

import dnautils

from immunedb.identification import add_sequences, AlignmentException
from immunedb.identification.vdj_sequence import VDJAlignment, VDJSequence
from immunedb.identification.genes import (CDR3_OFFSET, GeneName, JGermlines,
                                           VGermlines)
from immunedb.identification.identify import IdentificationProps
from immunedb.common.models import (DuplicateSequence, NoResult, Sample,
                                    Sequence, serialize_gaps)
from immunedb.util.funcs import format_ties, periodic_commit
import immunedb.util.lookups as lookups
from immunedb.util.log import logger


GAP_PLACEHOLDER = '.'


def get_fasta(sequences):
    return '\n'.join(['>{}\n{}'.format(k, v.replace('-', ''))
                     for k, v in sequences.iteritems()]) + '\n'


def gaps_before(gaps, pos):
    return sum((e[1] for e in gaps if e[0] < pos))


def gap_positions(seq, char='-'):
    gaps = []
    for diff in re.finditer('[{}]+'.format(char), seq):
        start, end = diff.span()
        gaps.append((start, end - start))
    return gaps


def build_index(germlines, path):
    sequences = get_fasta(germlines)
    cmd = ('bowtie2-build -c "{}" {}').format(sequences, path)
    proc = subprocess.Popen(shlex.split(cmd), stderr=subprocess.PIPE,
                            stdout=subprocess.PIPE)

    stdout, stderr = proc.communicate()
    return stdout


def align_reference(path, index, sequences, nproc):
    cmd = ('bowtie2 --local -x {} -U {} -f --no-unal --no-sq --no-head '
           '--very-sensitive-local -p {}').format(index, sequences, nproc)
    proc = subprocess.Popen(shlex.split(cmd),
                            stderr=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            cwd=path)

    stdout, stderr = proc.communicate()
    return stdout


def get_reader(output):
    fieldnames = [
        'seq_id',
        'flags',
        'reference',
        'ref_offset',
        'map_quality',
        'cigar',
        'mate_reference',
        'ref_mate_offset',
        'frag_size',
        'read_seq',
        'read_quality',
    ]
    return csv.DictReader(StringIO.StringIO(output), delimiter='\t',
                          fieldnames=fieldnames, restkey='optional')


def create_seqs(read_seq, ref_seq, cigar, ref_offset, min_size, **kwargs):
    ref_offset = max(0, ref_offset)
    full_ref = ref_seq[:]
    full_seq = read_seq[:]
    ref_seq = ref_seq[ref_offset:]

    final_seq = ['N' * ref_offset]
    final_ref = [ref_seq[:ref_offset]]
    skips = []

    ops = re.findall('(\d+)([A-Z])', cigar)
    for i, (cnt, op) in enumerate(ops):
        cnt = int(cnt)
        if op == 'M':
            final_seq.extend(read_seq[:cnt])
            final_ref.extend(ref_seq[:cnt])
            read_seq = read_seq[cnt:]
            ref_seq = ref_seq[cnt:]
        elif op == 'S':
            if i == len(ops) - 1:
                skips.append(read_seq[:cnt])
            read_seq = read_seq[cnt:]
        elif op == 'D':
            final_seq.extend(['-'] * cnt)
            final_ref.extend(ref_seq[:cnt])
            ref_seq = ref_seq[cnt:]
        elif op == 'I':
            final_ref.extend(['-'] * cnt)
            final_seq.extend(read_seq[:cnt])
            read_seq = read_seq[cnt:]
        else:
            raise Exception('Unknown opcode {}'.format(op))

    final_ref = ''.join(final_ref)
    final_seq = ''.join(final_seq)

    if len(final_ref) < min_size:
        ref_offset = len(final_ref) - final_ref.count('-')
        final_ref = ''.join([
            final_ref,
            full_ref[ref_offset:min_size]
        ])
        diff = len(final_ref) - len(final_seq)
        if diff > 0:
            seq_offset = len(final_seq) - final_seq.count('-')
            final_seq = ''.join([
                final_seq,
                full_seq[seq_offset:seq_offset + diff]
            ])
            if len(skips) > 0:
                skips[-1] = skips[-1][diff:]

    return (final_ref, final_seq, skips)


def add_imgt_gaps(imgt_germline, aligned_germline, sequence, seq_start):
    for pos, size in gap_positions(imgt_germline):
        pos += gaps_before(gap_positions(aligned_germline), pos)
        if pos < seq_start:
            seq_start += size
        aligned_germline = ''.join((
            aligned_germline[:pos],
            GAP_PLACEHOLDER * size,
            aligned_germline[pos:]
        ))
        sequence = ''.join((
            sequence[:pos],
            GAP_PLACEHOLDER * size,
            sequence[pos:]
        ))

    return (aligned_germline, sequence, seq_start)


def get_formatted_ties(genes):
    res = {}
    for ties, seq in genes.iteritems():
        res[format_ties(ties)] = seq
    return res


def process_sample(session, sample, indexes, temp, v_germlines, j_germlines,
                   nproc):
    indels = session.query(
        Sequence.ai,
        Sequence.seq_id,
        Sequence.sample_id,
        Sequence.sequence
    ).filter(
        Sequence.sample_id == sample.id,
        Sequence.probable_indel_or_misalign == 1
    ).order_by(Sequence.seq_id)
    # Get the sequences that were not identifiable
    noresults = session.query(NoResult).filter(
        NoResult.sample_id == sample.id).order_by(NoResult.seq_id)

    if indels.count() == 0 and noresults.count() == 0:
        logger.info('Sample {} has no indels or noresults'.format(
            sample.id))
        return
    logger.info('Sample {} has {} indels and {} noresults'.format(
                sample.id, indels.count(), noresults.count()))

    mut_bucket = v_germlines.mut_bucket(sample.v_ties_mutations)
    len_bucket = v_germlines.length_bucket(sample.v_ties_len)
    bucket = '{}_{}'.format(str(mut_bucket).replace('.', ''),
                            len_bucket)
    sample_v_germlines = get_formatted_ties(v_germlines.all_ties(
            sample.v_ties_len, sample.v_ties_mutations))
    sample_j_germlines = get_formatted_ties(j_germlines.all_ties(
        sample.v_ties_len, sample.v_ties_mutations))
    if bucket not in indexes:
        indexes.add(bucket)
        v_path = os.path.join(temp, 'v_genes_{}'.format(bucket))
        j_path = os.path.join(temp, 'j_genes_{}'.format(bucket))
        logger.info('Creating index for V-ties at {} length, {} '
                    'mutation'.format(len_bucket, mut_bucket))
        build_index(sample_v_germlines, v_path)
        build_index(sample_j_germlines, j_path)

    seq_path = os.path.join(temp, 'll_{}.fasta'.format(sample.id))
    with open(seq_path, 'w+') as fh:
        fh.write(get_fasta({'tp=Sequence|ai={}|sample_id={}|seq_id={}'.format(
                r.ai, r.sample_id, r.seq_id): r.sequence for r in indels}))
        fh.write(get_fasta({'tp=NoResult|pk={}|sample_id={}|seq_id={}'.format(
            r.pk, r.sample_id, r.seq_id): r.sequence for r in noresults}))

    alignments = {}
    logger.info('Running bowtie2 for V-gene sequences')
    for line in get_reader(align_reference(temp, 'v_genes_{}'.format(bucket),
                                           seq_path, nproc)):
        line['ref_offset'] = int(line['ref_offset']) - 1
        ref_gene = line['reference']
        try:
            ref, seq, rem_seqs = create_seqs(
                ref_seq=sample_v_germlines[ref_gene].replace('-', ''),
                min_size=CDR3_OFFSET, **line)
        except KeyError as e:
            logger.warning('bowtie got invalid V: ' + str(e))
            continue
        if len(rem_seqs) == 0:
            continue

        ref, seq, seq_start = add_imgt_gaps(sample_v_germlines[ref_gene], ref,
                                            seq, line['ref_offset'])
        if len(ref) < CDR3_OFFSET:
            continue
        alignments[line['seq_id']] = {
            'v_germline': ref,
            'v_gene': line['reference'],
            'seq_start': seq_start,
            'v_sequence': seq,
            'v_rem_seq': rem_seqs[-1],
            'cdr3_start': len(ref)
        }

    seq_path = os.path.join(temp, 'll_j_{}.fasta'.format(sample.id))
    with open(seq_path, 'w+') as fh:
        seqs = {k: v['v_rem_seq'] for k, v in alignments.iteritems() if
                len(v['v_rem_seq']) > 0}
        fh.write(get_fasta(seqs))

    tasks = []
    logger.info('Running bowtie2 for J-gene sequences')
    for line in get_reader(align_reference(temp, 'j_genes_{}'.format(bucket),
                                           seq_path, nproc)):
        line['ref_offset'] = int(line['ref_offset']) - 1
        ref_gene = line['reference']
        ref, seq, rem_seqs = create_seqs(
            ref_seq=sample_j_germlines[ref_gene].replace('-', ''),
            min_size=j_germlines.upstream_of_cdr3, **line)
        alignments[line['seq_id']]['j_gene'] = line['reference']

        full_seq = (alignments[line['seq_id']]['v_sequence'] +
                    alignments[line['seq_id']]['v_rem_seq'])
        if len(rem_seqs) > 0:
            full_seq = full_seq[:-len(rem_seqs[-1])]

        cdr3_end = len(full_seq)
        if len(ref) < j_germlines.upstream_of_cdr3:
            continue
        for i in range(j_germlines.upstream_of_cdr3):
            if ref[-i] != '-':
                cdr3_end -= 1
        alignments[line['seq_id']]['cdr3_end'] = cdr3_end

        cdr3_length = cdr3_end - alignments[line['seq_id']]['cdr3_start']

        full_germ = (alignments[line['seq_id']]['v_germline'] +
                     (GAP_PLACEHOLDER * cdr3_length))
        j_length = len(full_seq) - len(full_germ)
        if j_length <= 0 or cdr3_length <= 0:
            continue
        full_germ += ref[-j_length:]

        r_type, pk, sample_id, seq_id = [
            v.split('=', 1)[1] for v in line['seq_id'].split('|', 3)]
        insertions = gap_positions(full_germ)
        deletions = gap_positions(full_seq)

        alignment = VDJAlignment(
            VDJSequence(seq_id, full_seq.replace(GAP_PLACEHOLDER, '-'))
        )
        alignment.germline = full_germ.replace(GAP_PLACEHOLDER, '-')
        if len(alignment.germline) != len(alignment.sequence.sequence):
            continue
        alignment.v_gene.add(GeneName(alignments[line['seq_id']]['v_gene']))
        alignment.j_gene.add(GeneName(alignments[line['seq_id']]['j_gene']))
        alignment.seq_offset = alignments[line['seq_id']]['seq_start']
        # TODO: This should really look for a streak like in anchoring
        alignment.germline_cdr3 = '-' * cdr3_length
        gaps_in_seq = alignment.sequence.sequence[
            alignment.seq_start:alignments[line['seq_id']]['cdr3_start']
        ].count('-')
        alignment.v_length = (
            alignments[line['seq_id']]['cdr3_start'] -
            alignment.seq_offset
        ) - gaps_in_seq
        alignment.j_length = j_length
        alignment.v_mutation_fraction = 1 - (alignment.v_match /
                                             float(alignment.v_length))
        alignment.cdr3_start = alignments[line['seq_id']]['cdr3_start']
        alignment.cdr3_num_nts = cdr3_length
        alignment.post_cdr3_length = j_length
        alignment.insertions = insertions
        alignment.deletions = deletions
        alignment.locally_aligned = True

        tasks.append({
            'r_type': r_type,
            'pk': int(pk),
            'sample_id': int(sample_id),
            'alignment': alignment
        })
    return tasks


def add_sequences_from_sample(session, sample, sequences, props):
    logger.info('Adding {} corrected sequences to sample {}'.format(
        len(sequences), sample.id))
    for sequence in periodic_commit(session, sequences):
        alignment = sequence['alignment']
        try:
            try:
                props.validate(alignment)
            except AlignmentException:
                continue
            if sequence['r_type'] == 'NoResult':
                add_sequences(session, [alignment], sample,
                              error_action='raise')
                session.query(NoResult).filter(
                    NoResult.pk == sequence['pk']
                ).delete(synchronize_session=False)
            elif sequence['r_type'] == 'Sequence':
                fields = {
                    'partial': alignment.partial,

                    'probable_indel_or_misalign':
                        alignment.has_possible_indel,

                    'v_gene': format_ties(alignment.v_gene),
                    'j_gene': format_ties(alignment.j_gene),

                    'num_gaps': alignment.num_gaps,
                    'seq_start': alignment.seq_start,

                    'v_match': alignment.v_match,
                    'v_length': alignment.v_length,
                    'j_match': alignment.j_match,
                    'j_length': alignment.j_length,

                    'removed_prefix':
                        alignment.sequence.removed_prefix_sequence,
                    'removed_prefix_qual':
                        alignment.sequence.removed_prefix_quality,
                    'v_mutation_fraction': alignment.v_mutation_fraction,

                    'pre_cdr3_length': alignment.pre_cdr3_length,
                    'pre_cdr3_match': alignment.pre_cdr3_match,
                    'post_cdr3_length': alignment.post_cdr3_length,
                    'post_cdr3_match': alignment.post_cdr3_match,

                    'in_frame': alignment.in_frame,
                    'functional': alignment.functional,
                    'stop': alignment.stop,

                    'cdr3_nt': alignment.cdr3,
                    'cdr3_num_nts': len(alignment.cdr3),
                    'cdr3_aa': lookups.aas_from_nts(alignment.cdr3),

                    'sequence': str(alignment.sequence.sequence),
                    'quality': alignment.sequence.quality,

                    'locally_aligned': alignment.locally_aligned,
                    '_insertions': serialize_gaps(alignment.insertions),
                    '_deletions': serialize_gaps(alignment.deletions),

                    'germline': alignment.germline
                }
                # This line doesnt actually add anything to the DB, it's just
                # to validate the fields
                Sequence(**fields)

                session.query(Sequence).filter(
                    Sequence.ai == sequence['pk']
                ).update(fields, synchronize_session=False)
        except ValueError:
            continue


def remove_duplicates(session, sample):
    logger.info('Removing duplicates from sample {}'.format(sample.id))
    seqs = session.query(
        Sequence.ai, Sequence.seq_id, Sequence.v_gene, Sequence.j_gene,
        Sequence.cdr3_num_nts, Sequence.copy_number, Sequence.sequence
    ).filter(
        Sequence.locally_aligned.is_(True),
        Sequence.sample_id == sample.id
    ).order_by(Sequence.ai)

    for seq in seqs:
        potential_collapse = session.query(
            Sequence.seq_id, Sequence.sequence, Sequence.copy_number
        ).filter(
            Sequence.sample_id == sample.id,
            Sequence.v_gene == seq.v_gene,
            Sequence.j_gene == seq.j_gene,
            Sequence.cdr3_num_nts == seq.cdr3_num_nts,
        ).order_by(desc(Sequence.copy_number), Sequence.ai)

        for other_seq in potential_collapse:
            if (other_seq.seq_id == seq.seq_id or
                    len(other_seq.sequence) != len(seq.sequence)):
                continue

            if dnautils.equal(other_seq.sequence, seq.sequence):
                session.query(DuplicateSequence).filter(
                    DuplicateSequence.sample_id == sample.id,
                    DuplicateSequence.duplicate_seq_seq_id == seq.seq_id
                ).update({
                    'sample_id': sample.id,
                    'duplicate_seq_seq_id': other_seq.seq_id
                }, synchronize_session=False)
                session.add(DuplicateSequence(
                    seq_id=seq.seq_id,
                    duplicate_seq_seq_id=other_seq.seq_id,
                    sample_id=sample.id
                ))
                session.query(Sequence).filter(Sequence.ai == seq.ai).delete()
                break

    session.commit()


def run_fix_sequences(session, args):
    v_germlines = VGermlines(args.v_germlines)
    j_germlines = JGermlines(args.j_germlines, args.upstream_of_cdr3)

    indexes = set()
    props = IdentificationProps(**args.__dict__)
    samples = session.query(Sample)
    if args.sample_ids:
        samples = samples.filter(Sample.id.in_(args.sample_ids))
    for sample in samples:
        sequences = process_sample(session, sample, indexes, args.temp,
                                   v_germlines, j_germlines, args.nproc)
        add_sequences_from_sample(session, sample, sequences, props)
        remove_duplicates(session, sample)

    logger.info('Updating copy numbers')
    session.connection(mapper=Sequence).execute(text('''
        UPDATE
            sequences
        SET
            copy_number = 1 + (
                SELECT
                    COUNT(*)
                FROM
                    duplicate_sequences
                WHERE
                    duplicate_sequences.duplicate_seq_seq_id = sequences.seq_id
                    AND
                    duplicate_sequences.sample_id = sequences.sample_id
            )
    '''))



from collections import Counter
import itertools


def yield_limit(qry, pk_attr, maxrq=5000):
    firstid = None
    while True:
        q = qry
        if firstid is not None:
            q = qry.filter(pk_attr > firstid)
        rec = None
        for rec in q.order_by(pk_attr).limit(maxrq):
            yield rec
        if rec is None:
            break
        firstid = pk_attr.__get__(rec, pk_attr) if rec else None


def bulk_add(session, objs, chunk_size=100, flush=True):
    for i in xrange(0, len(objs), chunk_size):
        session.bulk_save_objects(objs[i:i + chunk_size])
        if flush:
            session.flush()


def flatten(iterable):
    return list(itertools.chain.from_iterable(iterable))


def consensus(strings):
    """Gets the unweighted consensus from a list of strings

    :param list strings: A set of equal-length strings.

    :returns: A consensus string
    :rtype: str

    """
    chrs = [Counter(chars).most_common(1)[0][0] for chars in zip(*strings)]
    return ''.join(chrs)


def get_regions(insertions):
    regions = [78, 36, 51, 30, 114]
    if insertions is not None and len(insertions) > 0:
        for pos, size in insertions:
            offset = 0
            for i, region_start in enumerate(regions):
                offset += region_start
                if pos < offset:
                    regions[i] += size
                    break

    return regions


def get_pos_region(regions, cdr3_len, pos):
    cdr3_start = sum(regions)
    j_start = cdr3_start + cdr3_len
    if pos >= j_start:
        return 'FR4'
    elif pos >= cdr3_start:
        return 'CDR3'

    total = 0
    for i, length in enumerate(regions):
        total += length
        if pos < total:
            rtype = 'FW' if i % 2 == 0 else 'CDR'
            rnum = (i // 2) + 1
            return '{}{}'.format(rtype, rnum)


def ord_to_quality(quality):
    if quality is None:
        return None
    return ''.join([' ' if q is None else chr(q + 33) for q in quality])


def periodic_commit(session, query, interval=10):
    for i, r in enumerate(query):
        if i > 0 and i % interval == 0:
            session.commit()
        yield r
    session.commit()


def get_or_create(session, model, **kwargs):
    """Gets or creates a record based on some kwargs search parameters"""
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance, False
    else:
        instance = model(**kwargs)
        session.add(instance)
        return instance, True


def format_ties(ties, strip_alleles=True):
    if ties is None:
        return None

    formatted = []
    for t in ties:
        prefix = t.prefix
        for e in t.name.split('|'):
            name = e.replace(prefix, '')
            if strip_alleles:
                name = name.split('*', 1)[0]
            formatted.append(name)
    return '{}{}'.format(prefix, '|'.join(sorted(set(formatted))))



import cStringIO as StringIO
import csv


class NestedCSVWriter(object):
    def __init__(self, headers, mapping=None, streaming=False):
        self._headers = headers
        self._mapping = mapping
        self._streaming = streaming
        self._out = StringIO.StringIO()
        self._csv = csv.DictWriter(self._out, fieldnames=headers)
        self._csv.writeheader()

    def add_row(self, row, write_default=False, default=None,
                write_if_stream=True):
        if write_default:
            write_dict = {k: default for k in self._headers}
        else:
            write_dict = {}
        for header in self._headers:
            if self._mapping is not None and header in self._mapping:
                write_dict[header] = self._mapping[header](row)
            elif header in row:
                write_dict[header] = row[header]
        return self.add_raw_row(write_dict, write_if_stream=write_if_stream)

    def add_raw_row(self, raw_dict, write_if_stream=True):
        self._csv.writerow(raw_dict)
        if self._streaming and write_if_stream:
            return self.get_value()

    def get_value(self):
        self._out.seek(0)
        v = self._out.read()
        if self._streaming:
            self._out.seek(0)
            self._out.truncate()
        return v



import csv
import dnautils
import os
import re

from immunedb.common.models import (CDR3_OFFSET, Sample, SampleMetadata, Study,
                                    Subject)
from immunedb.identification import (add_noresults_for_vdj, add_sequences,
                                     AlignmentException, get_common_seq)

from immunedb.identification.metadata import (MetadataException,
                                              parse_metadata, REQUIRED_FIELDS)
from immunedb.identification.genes import JGermlines, VGermlines, GeneName
from immunedb.identification.identify import IdentificationProps
from immunedb.identification.vdj_sequence import VDJAlignment, VDJSequence
import immunedb.util.funcs as funcs
from immunedb.util.log import logger


def create_alignment(seq, line, v_germlines, j_germlines):
    alignment = VDJAlignment(seq)
    alignment.v_gene = set([GeneName(g) for g in line['V_CALL'].split(',')])
    alignment.j_gene = set([GeneName(g) for g in line['J_CALL'].split(',')])
    alignment.cdr3_num_nts = int(line['JUNCTION_LENGTH'])
    alignment.v_length = int(line['V_SEQ_LENGTH'])
    alignment.seq_offset = re.match('[\-]*', alignment.sequence.sequence).end()

    # TODO: Calculate these
    alignment.v_length = CDR3_OFFSET - seq[:CDR3_OFFSET].count('-')
    alignment.j_length = j_germlines.upstream_of_cdr3

    germ_v = [v_germlines[g] for g in alignment.v_gene if g in v_germlines]
    germ_j = [j_germlines[g] for g in alignment.j_gene if g in j_germlines]
    if len(germ_v) == 0 or len(germ_j) == 0:
        raise AlignmentException('Missing germlines: V={} J={}'.format(
            ','.join([str(v) for v in alignment.v_gene]),
            ','.join([str(j) for j in alignment.j_gene])))

    germ_v = get_common_seq(germ_v)
    germ_j = get_common_seq(germ_j)

    alignment.germline_cdr3 = ''.join((
        germ_v,
        '-' * (len(alignment.sequence) - len(germ_v) - len(germ_j)),
        germ_j
    ))[CDR3_OFFSET:CDR3_OFFSET + alignment.cdr3_num_nts]

    alignment.germline = ''.join([
        germ_v[:CDR3_OFFSET],
        '-' * alignment.cdr3_num_nts,
        germ_j[-j_germlines.upstream_of_cdr3:]
    ])

    if len(alignment.germline) != len(alignment.sequence.sequence):
        raise AlignmentException('Sequence and germline differ in size')
    return alignment


def read_file(session, handle, sample, v_germlines, j_germlines, props):
    reader = csv.DictReader(handle, delimiter='\t')
    uniques = {}

    for line in reader:
        seq = VDJSequence(line['SEQUENCE_ID'],
                          line['SEQUENCE_IMGT'].replace('.', '-'))
        try:
            alignment = create_alignment(seq, line, v_germlines, j_germlines)
        except AlignmentException as e:
            add_noresults_for_vdj(session, seq, sample, str(e))
        for other in uniques.setdefault(len(alignment.sequence.sequence), []):
            if dnautils.equal(other.sequence.sequence,
                              alignment.sequence.sequence):
                other.sequence.ids.extend(alignment.sequence.ids)
                break
        else:
            uniques[len(alignment.sequence.sequence)].append(alignment)

    uniques = [s for k in sorted(uniques.keys()) for s in uniques[k]]
    lens = []
    muts = []
    for unique in uniques:
        try:
            props.validate(unique)
            add_sequences(session, [unique], sample)
            lens.append(unique.v_length)
            muts.append(unique.v_mutation_fraction)
        except AlignmentException as e:
            add_noresults_for_vdj(session, seq, sample, str(e))

    if len(lens) > 0:
        sample.v_ties_len = sum(lens) / float(len(lens))
        sample.v_ties_mutations = sum(muts) / float(len(muts))

    session.commit()


def create_sample(session, metadata):
    study, new = funcs.get_or_create(
        session, Study, name=metadata['study_name'])

    if new:
        logger.info('Created new study "{}"'.format(study.name))
        session.commit()

    sample, new = funcs.get_or_create(
        session, Sample, name=metadata['sample_name'], study=study)
    if new:
        logger.info('Created new sample "{}"'.format(sample.name))
        for key, value in metadata.iteritems():
            if key not in REQUIRED_FIELDS:
                session.add(SampleMetadata(
                    sample=sample,
                    key=key,
                    value=value
                ))

        subject, new = funcs.get_or_create(
            session, Subject, study=study,
            identifier=metadata['subject'])
        sample.subject = subject
        session.commit()
    else:
        logger.error(
            'Sample "{}" already exists'.format(metadata['sample_name']))
        return
    return sample


def run_import(session, args):
    v_germlines = VGermlines(args.v_germlines)
    j_germlines = JGermlines(args.j_germlines, args.upstream_of_cdr3,
                             args.anchor_len)

    meta_fn = args.metadata if args.metadata else os.path.join(
        args.sample_dir, 'metadata.tsv')

    if not os.path.isfile(meta_fn):
        logger.error('Metadata file not found.')
        return

    with open(meta_fn, 'rU') as fh:
        try:
            metadata = parse_metadata(session, fh, args.warn_existing,
                                      args.warn_missing, args.sample_dir)
        except MetadataException as ex:
            logger.error(ex.message)
            return

    props = IdentificationProps(**args.__dict__)
    for sample_name in sorted(metadata.keys()):
        sample = create_sample(session, metadata[sample_name])
        if sample:
            path = os.path.join(
                args.sample_dir, metadata[sample_name]['file_name'])
            with open(path) as fh:
                read_file(session, fh, sample, v_germlines, j_germlines, props)



import json
from functools import wraps
import re
import sys
import time

from sqlalchemy import distinct
from sqlalchemy.orm import scoped_session

import bottle
from bottle import response, request

try:
    from rollbar.contrib.bottle import RollbarBottleReporter
    ROLLBAR_SUPPORT = True
except ImportError:
    ROLLBAR_SUPPORT = False

from immunedb.exporting.clone_export import CloneExport
from immunedb.exporting.sequence_export import SequenceExport
from immunedb.exporting.mutation_export import MutationExporter
import immunedb.api.queries as queries
from immunedb.common.models import CloneStats
from immunedb.exporting.writers import (CLIPWriter, FASTAWriter, FASTQWriter,
                                        CSVWriter)
from immunedb.util.log import logger


class EnableCors(object):
    """A class to enable Cross-Origin Resource Sharing to facilitate AJAX
    requests.

    """
    name = 'enable_cors'
    api = 2

    def apply(self, fn, context):
        def _enable_cors(*args, **kwargs):
            response.headers['Access-Control-Allow-Origin'] = '*'
            response.headers['Access-Control-Allow-Methods'] = ('GET, POST')
            response.headers['Access-Control-Allow-Headers'] = (
                'Origin, '
                'Accept, Content-Type, '
                'X-Requested-With, X-CSRF-Token')

            if bottle.request.method != 'OPTIONS':
                return fn(*args, **kwargs)

        return _enable_cors


def create_response(j=None, code=200, ctype='application/json'):
    if j is None:
        bottle.response.status = code
        return
    bottle.response.content_type = ctype
    bottle.response.status = code
    if ctype == 'application/json':
        return json.dumps(j)
    return j


def get_paging():
    data = bottle.request.json or {}
    return map(int, (
        data.get('page', 1),
        data.get('per_page', 10)
    ))


def decode_run_length(encoding):
    ids = []
    offset = 1
    for match in re.finditer('(T|F)(\d+)', encoding.upper()):
        size = int(match.group(2))
        if match.group(1) == 'T':
            ids.extend(range(offset, offset + size))
        offset += size
    return ids


def set_download(exp_type, ext):
    time_str = time.strftime('%Y-%m-%d-%H-%M')
    response.headers['Content-Disposition'] = 'attachment;filename={}'.format(
        '{}_{}.{}'.format(exp_type, time_str, ext))


def create_app(session_maker, allow_shutdown=False):
    app = bottle.Bottle()
    app.install(EnableCors())

    def with_session(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            session = scoped_session(session_maker)()
            try:
                return f(session, *args, **kwargs)
            except Exception:
                raise
            finally:
                session.close()

        return _wrapper

    @app.route('/samples/list', method=['POST', 'OPTIONS'])
    @with_session
    def samples_list(session):
        return create_response(queries.get_samples(session))

    @app.route('/sequences/list', method=['POST', 'OPTIONS'])
    @with_session
    def sequences_list(session):
        fields = bottle.request.json or {}
        return create_response(queries.get_sequences(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            fields.get('subject_id', None),
            get_paging()))

    @app.route('/sequence/<sample_id>/<seq_id>', method=['POST', 'OPTIONS'])
    @with_session
    def sequence(session, sample_id, seq_id):
        return create_response(queries.get_sequence(session, sample_id,
                                                    seq_id))

    @app.route('/clones/list', method=['POST', 'OPTIONS'])
    @app.route('/clones/list/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clones_list(session, subject_id=None):
        fields = bottle.request.json or {}
        return create_response(queries.get_clones(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            subject_id,
            get_paging()))

    @app.route('/clone/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone(session, clone_id):
        return create_response(queries.get_clone(session, clone_id))

    @app.route('/clone/sequences/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_sequences(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_sequences(
            session, clone_id, fields.get('get_collapse', False), get_paging()
        ))

    @app.route('/clone/mutations/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_mutations(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_mutations(
            session,
            clone_id,
            fields.get('type', 'percent'),
            int(fields.get('value', 0))
        ))

    @app.route('/clone/lineage/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_lineage(session, clone_id):
        return create_response(queries.get_clone_tree(session, clone_id))

    @app.route('/clone/pressure/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_pressure(session, clone_id):
        return create_response(
            queries.get_selection_pressure(session, clone_id))

    @app.route('/samples/analyze/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def analyze_samples(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(
            queries.analyze_samples(
                session,
                decode_run_length(sample_encoding),
                fields.get('filter_type', 'unique_multiple'),
                fields.get('include_outliers', True),
                fields.get('include_partials', True),
                fields.get('percentages', False),
                fields.get('grouping', 'name'),
            )
        )

    @app.route('/samples/overlap/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def overlap(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_overlap(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'clones_all'),
            get_paging())
        )

    @app.route('/samples/v_usage/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def v_usage(session, sample_encoding):
        fields = bottle.request.json or {}
        data, x_categories, totals, prefix = queries.get_v_usage(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'unique_multiple'),
            fields.get('include_outliers', True),
            fields.get('include_partials', True),
            fields.get('grouping', 'name'),
            fields.get('by_family', False)
        )

        x_categories.sort()
        y_categories = sorted(data.keys())
        array = []
        min_v = 100
        max_v = 0
        for i, x in enumerate(x_categories):
            for j, y in enumerate(y_categories):
                usage_for_y = data[y]
                if x in usage_for_y:
                    min_v = min(min_v, usage_for_y[x])
                    max_v = max(max_v, usage_for_y[x])
                    array.append([i, j, usage_for_y[x]])
                else:
                    array.append([i, j, 0])

        return create_response({
            'x_categories': ['{}{}'.format(prefix, e) for e in x_categories],
            'y_categories': y_categories,
            'totals': totals,
            'data': array,
            'min': min_v,
            'max': max_v
        })

    @app.route('/subjects/list', method=['POST', 'OPTIONS'])
    @with_session
    def subjects_list(session):
        return create_response(queries.get_all_subjects(session, get_paging()))

    @app.route('/subject/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def subject(session, subject_id):
        return create_response(queries.get_subject(session, subject_id))

    @app.route('/export/sequences/<from_type>/<encoding>', method=['POST'])
    @with_session
    def export_sequences(session, from_type, encoding):
        fields = request.forms or {}
        eformat = fields.get('format')

        set_download('sequences', eformat.split('-', 1)[0])

        if eformat == 'csv':
            writer = CSVWriter()
        elif eformat.startswith('fasta'):
            writer = FASTAWriter(replaced_sequences='fill' in eformat)
        elif eformat.startswith('fastq'):
            writer = FASTQWriter(replaced_sequences='fill' in eformat)
        elif eformat.startswith('clip'):
            writer = CLIPWriter(replaced_sequences='fill' in eformat)

        if from_type == 'sample':
            ids = decode_run_length(encoding)
        else:
            ids = [int(encoding)]

        export = SequenceExport(
            session, writer, from_type, ids,
            selected_fields=fields.get('fields').split(','),
            subject_uniques=fields.get('subject_uniques', False),
            only_with_clones=fields.get('only_with_clones', False)
        )
        for line in export.get_data():
            yield line

    @app.route('/export/clones/<from_type>/<encoding>', method=['POST'])
    @with_session
    def export_clones(session, from_type, encoding):
        fields = request.forms or {}

        set_download('clones', 'csv')

        if from_type == 'sample':
            ids = decode_run_length(encoding)
        else:
            ids = [int(encoding)]

        export = CloneExport(
            session, from_type, ids,
            fields.get('fields').split(',')
        )
        for line in export.get_data():
            yield line

    @app.route('/export/mutations/<from_type>/<encoding>', method=['POST'])
    @with_session
    def export_mutations(session, from_type, encoding):
        fields = request.forms or {}

        set_download('mutations', 'csv')

        if from_type == 'sample':
            query = session.query(
                distinct(CloneStats.clone_id).label('clone_id')
            ).filter(
                CloneStats.sample_id.in_(decode_run_length(encoding))
            )
            clone_ids = map(lambda r: r.clone_id, query.all())
        else:
            clone_ids = [int(encoding)]

        export = MutationExporter(
            session, clone_ids,
            decode_run_length(encoding)
            if fields.get('only_sample_rows') else None,
            fields.get('thresh_type', 'sequences'),
            int(fields.get('thresh_value', 0))
        )

        for line in export.get_data():
            yield line

    @app.route('/shutdown', method=['POST'])
    def shutdown():
        if allow_shutdown:
            logger.warning('Shutting down from remote request')
            sys.exit()
        return create_response(code=404)

    return app


def run_rest_service(session_maker, args):
    if args.rollbar_token:
        if not ROLLBAR_SUPPORT:
            logger.error('Rollbar is not installed')
            return
        rbr = RollbarBottleReporter(
            access_token=args.rollbar_token,
            environment=args.rollbar_env)
        bottle.install(rbr)

    app = create_app(session_maker, args.allow_shutdown)
    if args.debug:
        app.catchall = False
    app.run(host='0.0.0.0', port=args.port, server='gevent', debug=args.debug)



from collections import OrderedDict

from sqlalchemy import desc
from sqlalchemy.sql import text

import dnautils
from immunedb.common.models import (CDR3_OFFSET, Clone, Sequence,
                                    SequenceCollapse, Subject)
import immunedb.common.modification_log as mod_log
import immunedb.common.config as config
from immunedb.trees import cut_tree, get_seq_pks, PhylogeneticTree
import immunedb.util.concurrent as concurrent
import immunedb.util.funcs as funcs
import immunedb.util.lookups as lookups
from immunedb.util.log import logger


def generate_consensus(session, clone_ids):
    """Generates consensus CDR3s for clones.

    :param Session session: The database session
    :param list clone_ids: The list of clone IDs to assign to groups

    """

    if len(clone_ids) == 0:
        return
    for clone in funcs.periodic_commit(
            session,
            session.query(Clone).filter(Clone.id.in_(clone_ids)),
            interval=1000):
        seqs = session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.clone_id == clone.id,
            SequenceCollapse.copy_number_in_subject > 0
        ).all()
        clone.cdr3_nt = funcs.consensus([s.cdr3_nt for s in seqs])
        clone.cdr3_aa = lookups.aas_from_nts(clone.cdr3_nt)

        clone.germline = generate_germline(session, seqs, clone)

    session.commit()


def generate_germline(session, seqs, clone):
    rep_seq = seqs[0]
    cdr3_start_pos = sum(rep_seq.regions[:5])
    germline = rep_seq.germline[:cdr3_start_pos]
    germline += '-' * clone.cdr3_num_nts
    clone.functional = (
        len(germline) % 3 == 0 and
        not lookups.has_stop(germline)
    )

    j_region = rep_seq.germline[cdr3_start_pos + rep_seq.cdr3_num_nts:]
    germline += j_region

    return germline


def push_clone_ids(session):
    session.connection(mapper=Sequence).execute(text('''
        UPDATE
            sequences AS s
        JOIN sequence_collapse AS c
            ON s.sample_id=c.sample_id AND s.ai=c.seq_ai
        JOIN sequences as s2
            ON c.collapse_to_subject_seq_ai=s2.ai
        SET s.clone_id=s2.clone_id
        WHERE s.seq_id!=s2.seq_id
    '''))


def similar_to_all(seq, rest, min_similarity):
    """Determines if the string ``seq`` is at least ``min_similarity``
    similar to the list of strings ``rest``.

    :param str seq: The string to compare
    :param list rest: The list of strings to compare to
    :param int min_similarity: Minimum fraction to be considered similar

    :returns: If ``seq`` is similar to every sequence in ``rest``
    :rtype: bool

    """
    for comp_seq in rest:
        dist = dnautils.hamming(
            comp_seq.cdr3_aa.replace('X', '-'),
            seq.cdr3_aa.replace('X', '-')
        )
        sim_frac = 1 - dist / float(len(comp_seq.cdr3_aa))
        if sim_frac < min_similarity:
            return False
    return True


def can_subclone(sub_seqs, parent_seqs, min_similarity):
    for seq in sub_seqs:
        if not similar_to_all(seq, parent_seqs, min_similarity):
            return False
    return True


class ClonalWorker(concurrent.Worker):
    defaults = {
        # common
        'include_indels': False,
        'exclude_partials': False,
        'min_identity': 0,
        'min_copy': 2,
        'max_padding': None,

        # similarity
        'min_similarity': .85,

        # lineage
        'mut_cutoff': 4,
        'min_mut_occurrence': 2,
        'min_mut_samples': 1,
        'min_seq_instances': 1,
    }

    def __init__(self, session, **kwargs):
        self.session = session
        for prop, default in self.defaults.iteritems():
            setattr(self, prop, kwargs.get(prop, default))
        for prop, value in kwargs.iteritems():
            if prop not in self.defaults:
                setattr(self, prop, value)
        self._tasks = 0

    def get_bucket_seqs(self, bucket, sort):
        query = self.session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.subject_id == bucket.subject_id,
            Sequence.v_gene == bucket.v_gene,
            Sequence.j_gene == bucket.j_gene,
            Sequence.cdr3_num_nts == bucket.cdr3_num_nts,
            Sequence._insertions == bucket._insertions,
            Sequence._deletions == bucket._deletions,

            SequenceCollapse.copy_number_in_subject >= self.min_copy,
        )
        if sort:
            query = query.order_by(
                desc(SequenceCollapse.copy_number_in_subject),
                Sequence.ai
            )
        if self.min_identity > 0:
            query = query.filter(
                Sequence.v_match / Sequence.v_length >= self.min_identity
            )
        if self.min_seq_instances > 1:
            query = query.filter(
                SequenceCollapse.instances_in_subject >= self.min_seq_instances
            )
        if not self.include_indels:
            query = query.filter(Sequence.probable_indel_or_misalign == 0)
        if self.exclude_partials:
            query = query.filter(Sequence.partial == 0)
        if self.max_padding is not None:
            query = query.filter(Sequence.seq_start <= self.max_padding)
        return query

    def do_task(self, bucket):
        self.run_bucket(bucket)
        self._tasks += 1
        if self._tasks % 100 == 0:
            self.session.commit()
            self.info('Collapsed {} buckets'.format(self._tasks))

    def cleanup(self):
        self.session.commit()
        self.session.close()


class LineageClonalWorker(ClonalWorker):
    def run_bucket(self, bucket):
        updates = []
        consensus_needed = set([])

        seqs = self.get_bucket_seqs(bucket, sort=False)
        if seqs.count() > 0:
            cdr3_start = CDR3_OFFSET
            if bucket._insertions:
                cdr3_start += sum(
                    (i[1] for i in bucket._insertions)
                )
            germline = seqs[0].germline
            germline = ''.join((
                germline[:cdr3_start],
                '-' * bucket.cdr3_num_nts,
                germline[cdr3_start + bucket.cdr3_num_nts:]
            ))
            phylo = PhylogeneticTree(
                germline, seqs,
                min_mut_occurrence=self.min_mut_occurrence,
                min_mut_samples=self.min_mut_samples,
            )
            phylo.run(self.session, self.clearcut_path)

            for subtree in cut_tree(phylo.tree, self.mut_cutoff):
                new_clone = Clone(
                      subject_id=bucket.subject_id,
                      v_gene=bucket.v_gene,
                      j_gene=bucket.j_gene,
                      cdr3_num_nts=bucket.cdr3_num_nts,
                      _insertions=bucket._insertions,
                      _deletions=bucket._deletions
                )
                self.session.add(new_clone)
                self.session.flush()
                consensus_needed.add(new_clone.id)
                updates.extend([{
                    'sample_id': s[0],
                    'ai': s[1],
                    'clone_id': new_clone.id
                } for s in get_seq_pks(subtree)])

        if len(updates) > 0:
            self.session.bulk_update_mappings(Sequence, updates)
        generate_consensus(self.session, consensus_needed)


class TCellClonalWorker(ClonalWorker):
    def run_bucket(self, bucket):
        updates = []
        clones = OrderedDict()
        consensus_needed = set([])

        for seq in self.get_bucket_seqs(bucket, sort=False):
            key = (seq.v_gene, seq.j_gene, seq.cdr3_nt)
            if key in clones:
                clone = clones[key]
            else:
                for test_clone in clones.values():
                    same_bin = (test_clone.v_gene == key[0] and
                                test_clone.j_gene == key[1] and
                                test_clone.cdr3_num_nts == len(key[2]))
                    if same_bin and dnautils.equal(test_clone.cdr3_nt, key[2]):
                        clone = test_clone
                        break
                else:
                    new_clone = Clone(subject_id=seq.subject_id,
                                      v_gene=seq.v_gene,
                                      j_gene=seq.j_gene,
                                      cdr3_nt=seq.cdr3_nt,
                                      cdr3_num_nts=seq.cdr3_num_nts,
                                      _insertions=seq._insertions,
                                      _deletions=seq._deletions)
                    clones[key] = new_clone
                    self.session.add(new_clone)
                    self.session.flush()
                    clone = new_clone
                    consensus_needed.add(new_clone.id)
            updates.append({
                'sample_id': seq.sample_id,
                'ai': seq.ai,
                'clone_id': clone.id
            })

        if len(updates) > 0:
            self.session.bulk_update_mappings(Sequence, updates)
        generate_consensus(self.session, consensus_needed)


class SimilarityClonalWorker(ClonalWorker):
    def run_bucket(self, bucket):
        clones = OrderedDict()
        consensus_needed = set([])
        query = self.get_bucket_seqs(bucket, sort=True)

        if query.count() > 0:
            for seq in query:
                if seq.clone_id not in clones:
                    clones[seq.clone_id] = []
                clones[seq.clone_id].append(seq)
            if None in clones:
                for seq_to_add in clones[None]:
                    for clone_id, existing_seqs in clones.iteritems():
                        if clone_id is None:
                            continue
                        if similar_to_all(seq_to_add, existing_seqs,
                                          self.min_similarity):
                            existing_seqs.append(seq_to_add)
                            break
                    else:
                        new_clone = Clone(subject_id=seq.subject_id,
                                          v_gene=seq.v_gene,
                                          j_gene=seq.j_gene,
                                          cdr3_num_nts=seq.cdr3_num_nts,
                                          _insertions=seq._insertions,
                                          _deletions=seq._deletions)
                        self.session.add(new_clone)
                        self.session.flush()
                        clones[new_clone.id] = [seq_to_add]
                del clones[None]

            for clone_id, seqs in clones.iteritems():
                to_update = [
                    {
                        'sample_id': s.sample_id,
                        'ai': s.ai,
                        'clone_id': clone_id
                    } for s in seqs if s.clone_id is None
                ]
                if len(to_update) > 0:
                    self.session.bulk_update_mappings(Sequence, to_update)
                    consensus_needed.add(clone_id)
        generate_consensus(self.session, consensus_needed)


class SubcloneWorker(concurrent.Worker):
    def __init__(self, session, min_similarity):
        self.session = session
        self.min_similarity = min_similarity

    def do_task(self, bucket):
        clones = self.session.query(Clone).filter(
            Clone.subject_id == bucket.subject_id,
            Clone.v_gene == bucket.v_gene,
            Clone.j_gene == bucket.j_gene,
            Clone.cdr3_num_nts == bucket.cdr3_num_nts,
        ).all()

        if len(clones) == 0:
            return
        # The clones with indels are the only ones which can be subclones
        parent_clones = set([c for c in clones if len(c.insertions) == 0 and
                             len(c.deletions) == 0])
        potential_subclones = [c for c in clones if c not in parent_clones and
                               c.parent_id is None]
        self.info('Bucket {} has {} clones; parents={}, subs={}'.format(
            bucket, len(clones), len(parent_clones), len(potential_subclones)))

        for subclone in potential_subclones:
            for parent in parent_clones:
                if can_subclone(subclone.sequences, parent.sequences,
                                self.min_similarity):
                    subclone.parent = parent
                    break
        self.session.commit()

    def cleanup(self):
        self.session.commit()
        self.session.close()


def run_subclones(session, subject_ids, args):
    tasks = concurrent.TaskQueue()
    for subject_id in subject_ids:
        logger.info('Generating subclone task queue for subject {}'.format(
            subject_id))
        buckets = session.query(
            Clone.subject_id, Clone.v_gene, Clone.j_gene, Clone.cdr3_num_nts
        ).filter(
            Clone.subject_id == subject_id
        ).group_by(
            Clone.subject_id, Clone.v_gene, Clone.j_gene, Clone.cdr3_num_nts
        )
        for bucket in buckets:
            tasks.add_task(bucket)

    logger.info('Generated {} total subclone tasks'.format(tasks.num_tasks()))
    for i in range(0, min(tasks.num_tasks(), args.nproc)):
        tasks.add_worker(SubcloneWorker(config.init_db(args.db_config),
                                        args.similarity / 100.0))
    tasks.start()


def run_clones(session, args):
    """Runs the clone-assignment pipeline stage.

    :param Session session: The database session
    :param Namespace args: The arguments passed to the command

    """
    if args.subject_ids is None:
        subject_ids = map(lambda s: s.id, session.query(Subject.id).all())
    else:
        subject_ids = args.subject_ids
    mod_log.make_mod('clones', session=session, commit=True, info=vars(args))

    if args.regen:
        logger.info('Deleting existing clones')
        session.query(Clone).filter(
            Clone.subject_id.in_(subject_ids)
        ).delete(synchronize_session=False)
        session.commit()

    tasks = concurrent.TaskQueue()
    for subject_id in subject_ids:
        logger.info('Generating task queue for subject {}'.format(
            subject_id))
        buckets = session.query(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts, Sequence._insertions,
            Sequence._deletions
        ).filter(
            Sequence.subject_id == subject_id,
            Sequence.clone_id.is_(None)
        ).group_by(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts, Sequence._insertions,
            Sequence._deletions
        )
        for bucket in buckets:
            tasks.add_task(bucket)

    logger.info('Generated {} total tasks'.format(tasks.num_tasks()))

    methods = {
        'similarity': SimilarityClonalWorker,
        'tcells': TCellClonalWorker,
        'lineage': LineageClonalWorker,
    }
    for i in range(0, min(tasks.num_tasks(), args.nproc)):
        worker = methods[args.method](
            config.init_db(args.db_config), **args.__dict__
        )
        tasks.add_worker(worker)
    tasks.start()

    if args.subclones:
        run_subclones(session, subject_ids, args)
    else:
        logger.info('Skipping subclones')

    push_clone_ids(session)
    session.commit()



import itertools


def aas_from_nts(nts, replace_unknowns='X'):
    """Returns the amino acids for a given nuceotide string"""
    aas = ''
    for i in range(0, len(nts), 3):
        codon = str(nts[i:i+3])
        if len(codon) < 3:
            break
        aas += aa_from_codon(codon, replace_unknowns)
    return aas


def aa_from_codon(codon, ret_on_unknown=None):
    """Looks up an amino acid from a codon, or returns None"""
    if codon.upper() in _aa_lookup:
        return _aa_lookup[codon.upper()]
    return ret_on_unknown


def are_conserved_aas(aa1, aa2):
    """Determines if two amino acids are conserved based on the paper
    'Structural Determinants in the Sequences of Immunoglobulin Variable
    Domain'"""
    groups = [
        'WFILMVC',
        'PYTHSAG',
        'KEQNRD'
    ]
    for g in groups:
        if aa1.upper() in g and aa2.upper() in g:
            return True
    return False


def aa_to_all_nts(aas):
    nts = []
    for aa in aas:
        nts.append(_nt_lookup[aa])
    return map(lambda arr: ''.join(arr), itertools.product(*nts))


def has_stop(seq):
    for i in range(0, len(seq), 3):
        if seq[i:i+3] in ['TAG', 'TAA', 'TGA']:
            return True
    return False


_aa_lookup = {'TTT': 'F', 'TTC': 'F', 'TTA': 'L', 'TTG': 'L', 'TCT': 'S',
              'TCC': 'S', 'TCA': 'S', 'TCG': 'S', 'TAT': 'Y', 'TAC': 'Y',
              'TAA': '*', 'TAG': '*', 'TGT': 'C', 'TGC': 'C', 'TGA': '*',
              'TGG': 'W', 'CTT': 'L', 'CTC': 'L', 'CTA': 'L', 'CTG': 'L',
              'CCT': 'P', 'CCC': 'P', 'CCA': 'P', 'CCG': 'P', 'CAT': 'H',
              'CAC': 'H', 'CAA': 'Q', 'CAG': 'Q', 'CGT': 'R', 'CGC': 'R',
              'CGA': 'R', 'CGG': 'R', 'ATT': 'I', 'ATC': 'I', 'ATA': 'I',
              'ATG': 'M', 'ACT': 'T', 'ACC': 'T', 'ACA': 'T', 'ACG': 'T',
              'AAT': 'N', 'AAC': 'N', 'AAA': 'K', 'AAG': 'K', 'AGT': 'S',
              'AGC': 'S', 'AGA': 'R', 'AGG': 'R', 'GTT': 'V', 'GTC': 'V',
              'GTA': 'V', 'GTG': 'V', 'GCT': 'A', 'GCC': 'A', 'GCA': 'A',
              'GCG': 'A', 'GAT': 'D', 'GAC': 'D', 'GAA': 'E', 'GAG': 'E',
              'GGT': 'G', 'GGC': 'G', 'GGA': 'G', 'GGG': 'G'}


_nt_lookup = {
    'A': ['GCA', 'GCC', 'GCG', 'GCT'],
    'C': ['TGT', 'TGC'],
    'E': ['GAG', 'GAA'],
    'D': ['GAT', 'GAC'],
    'G': ['GGT', 'GGG', 'GGA', 'GGC'],
    'F': ['TTT', 'TTC'],
    'I': ['ATC', 'ATA', 'ATT'],
    'H': ['CAT', 'CAC'],
    'K': ['AAG', 'AAA'],
    '*': ['TAG', 'TAA', 'TGA'],
    'M': ['ATG'],
    'L': ['CTT', 'CTG', 'CTA', 'CTC', 'TTA', 'TTG'],
    'N': ['AAC', 'AAT'],
    'Q': ['CAA', 'CAG'],
    'P': ['CCT', 'CCG', 'CCA', 'CCC'],
    'S': ['AGC', 'AGT', 'TCT', 'TCG', 'TCC', 'TCA'],
    'R': ['AGG', 'AGA', 'CGA', 'CGG', 'CGT', 'CGC'],
    'T': ['ACA', 'ACG', 'ACT', 'ACC'],
    'W': ['TGG'],
    'V': ['GTA', 'GTC', 'GTG', 'GTT'],
    'Y': ['TAT', 'TAC']
}



import re

from Bio.Seq import Seq

import dnautils
from immunedb.common.models import CDR3_OFFSET
import immunedb.util.lookups as lookups


class VDJSequence(object):
    def __init__(self, ids, sequence, quality=None):
        if quality and len(sequence) != len(quality):
            raise ValueError('Sequence and quality must be the same length')
        if not all([c in 'ATCGN-' for c in sequence]):
            raise ValueError('Invalid characters in sequence: {}'.format(
                sequence))

        self.ids = [ids] if type(ids) == str else ids
        self.orig_sequence = sequence[:]
        self.orig_quality = quality[:] if quality else None
        self._sequence = sequence
        self._quality = quality
        self._removed_prefix_sequence = ''
        self._removed_prefix_quality = ''

    @property
    def sequence(self):
        return self._sequence

    @property
    def quality(self):
        return self._quality

    @property
    def removed_prefix_sequence(self):
        return self._removed_prefix_sequence

    @property
    def removed_prefix_quality(self):
        return self._removed_prefix_quality

    def reverse_complement(self, in_place=False):
        if not in_place:
            return VDJSequence(
                self.ids,
                str(Seq(self._sequence).reverse_complement()),
                self._quality[::-1] if self._quality else None
            )
        self._sequence = str(Seq(self._sequence).reverse_complement())
        self._quality = self._quality[::-1]

    def pad(self, count):
        self._sequence = ('N' * count) + self._sequence
        if self._quality:
            self._quality = (' ' * count) + self._quality

    def pad_right(self, count):
        self._sequence += 'N' * count
        if self._quality:
            self._quality += ' ' * count

    def remove_prefix(self, count):
        self._removed_prefix_sequence = self._sequence[:count]
        self._sequence = self._sequence[count:]
        if self._quality:
            self._removed_prefix_quality = self._sequence[:count]
            self._quality = self._quality[count:]

    def trim(self, count):
        new_prefix = ''.join([
            c if c == '-' else 'N' for c in self._sequence[:count]
        ])
        self._sequence = new_prefix + self._sequence[count:]
        if self._quality:
            self._quality = (' ' * count) + self._quality[count:]

    def trim_right(self, count):
        self._sequence = self._sequence[:count]
        if self._quality:
            self._quality = self._quality[:count]

    def add_gap(self, pos, char='-'):
        self._sequence = self._sequence[:pos] + char + self._sequence[pos:]
        if self._quality:
            self._quality = self._quality[:pos] + ' ' + self._quality[pos:]

    def rfind(self, seq):
        return self._sequence.rfind(seq)

    def __getitem__(self, key):
        return self._sequence[key]

    def __setitem__(self, key, value):
        self._sequence[key] = value

    def __len__(self):
        return len(self._sequence)


class VDJAlignment(object):
    INDEL_WINDOW = 30
    INDEL_MISMATCH_THRESHOLD = .6

    def __init__(self, sequence):
        self.sequence = sequence
        self.germline = None
        self.j_gene = set()
        self.v_gene = set()

        self.locally_aligned = False
        self.seq_offset = 0
        self.v_length = 0
        self.j_length = 0
        self.v_mutation_fraction = 0
        self.cdr3_start = CDR3_OFFSET
        self.cdr3_num_nts = 0
        self.germline_cdr3 = None
        self.post_cdr3_length = 0
        self.insertions = set([])
        self.deletions = set([])

    @property
    def filled_germline(self):
        return ''.join((
            self.germline[:self.cdr3_start],
            self.germline_cdr3,
            self.germline[self.cdr3_start + self.cdr3_num_nts:]
        ))

    @property
    def seq_start(self):
        return max(0, self.seq_offset)

    @property
    def num_gaps(self):
        return self.sequence[self.seq_start:self.cdr3_start].count('-')

    @property
    def cdr3(self):
        return self.sequence[self.cdr3_start:self.cdr3_start +
                             self.cdr3_num_nts]

    @property
    def partial(self):
        return self.seq_start > 0

    @property
    def in_frame(self):
        return len(self.cdr3) % 3 == 0 and self.cdr3_start % 3 == 0

    @property
    def stop(self):
        return lookups.has_stop(self.sequence)

    @property
    def functional(self):
        return self.in_frame and not self.stop

    @property
    def v_match(self):
        start = self.seq_start
        end = start + self.v_length + self.num_gaps

        return self.v_length - dnautils.hamming(
            self.filled_germline[start:end],
            self.sequence[start:end]
        )

    @property
    def j_match(self):
        return self.j_length - dnautils.hamming(
            self.filled_germline[-self.j_length:],
            self.sequence[-self.j_length:]
        )

    @property
    def pre_cdr3_length(self):
        return self.cdr3_start - self.seq_start - self.num_gaps

    @property
    def pre_cdr3_match(self):
        start = self.seq_start + self.num_gaps
        end = self.cdr3_start

        return self.pre_cdr3_length - dnautils.hamming(
            self.germline[start:end],
            self.sequence[start:end]
        )

    @property
    def post_cdr3_match(self):
        return self.post_cdr3_length - dnautils.hamming(
            self.germline[-self.post_cdr3_length:],
            self.sequence[-self.post_cdr3_length:]
        )

    @property
    def has_possible_indel(self):
        # Start comparison on first full AA to the INDEL_WINDOW or CDR3,
        # whichever comes first
        start = re.search('[ATCG]', self.sequence.sequence).start()
        germ = self.germline[start:self.cdr3_start]
        seq = self.sequence[start:self.cdr3_start]

        for i in range(0, len(germ) - self.INDEL_WINDOW + 1):
            dist = dnautils.hamming(germ[i:i+self.INDEL_WINDOW],
                                    seq[i:i+self.INDEL_WINDOW])
            if dist >= self.INDEL_MISMATCH_THRESHOLD * self.INDEL_WINDOW:
                return True

        return False

    def trim_to(self, count):
        old_pad = self.seq_start - self.sequence[:self.seq_start].count('-')

        self.sequence.trim(count)
        self.seq_offset = re.match('[N\-]*', self.sequence.sequence).span()[1]

        new_pad = self.sequence[:self.seq_start].count('-')
        self.v_length = self.v_length - self.seq_start + old_pad + new_pad



import logging

logger = logging.getLogger('immunedb')
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(
    logging.Formatter('%(asctime)s [%(levelname)s] %(message)s',
                      datefmt='%Y-%m-%d %H:%M:%S')
)
logger.addHandler(handler)



class Exporter(object):
    def __init__(self, session, rtype, rids, export_fields, selected_fields):
        self.session = session
        self.rtype = rtype
        self.rids = rids
        self.selected_fields = selected_fields
        self.export_fields = export_fields

    def get_selected_data(self, seq):
        """Gets the data specified by ``selected_fields`` for the sequence
        ``seq`` while overriding values in ``overrides`` if they exist.

        :param Sequence seq: The sequence from which to gather fields

        :returns: A list of ``(name, value)`` tuples with the selected data
        :rtype: list

        """
        return {
            name: accessor(seq) for name, accessor in
            self.export_fields.iteritems() if name in self.selected_fields
        }



from setuptools import setup, Extension

dnautils = Extension('dnautils', sources=['lib/dnautils.c'],
                     extra_compile_args=['-std=c99'])

setup(
    name='ImmuneDB',
    version='0.23.0',
    author='Aaron M. Rosenfeld',
    author_email='ar374@drexel.edu',
    url='https://github.com/arosenfeld/immunedb',
    packages=[
        'immunedb',
        'immunedb.aggregation',
        'immunedb.api',
        'immunedb.common',
        'immunedb.exporting',
        'immunedb.identification',
        'immunedb.importing',
        'immunedb.trees',
        'immunedb.util',
    ],
    scripts=[
        'bin/immunedb_admin',
        'bin/immunedb_clones',
        'bin/immunedb_clone_import',
        'bin/immunedb_clone_stats',
        'bin/immunedb_clone_pressure',
        'bin/immunedb_clone_trees',
        'bin/immunedb_collapse',
        'bin/immunedb_export',
        'bin/immunedb_local_align',
        'bin/immunedb_identify',
        'bin/immunedb_import',
        'bin/immunedb_metadata',
        'bin/immunedb_rest',
        'bin/immunedb_sample_stats',
        'bin/immunedb_sql',
        'bin/immunedb_genotype',
        'bin/run_tigger'
    ],
    install_requires=[
        'gevent',
        'sqlalchemy',
        'biopython',
        'bottle',
        'ete2',
        'numpy',
        'pygtrie',
        'PyMySQL',
    ],
    ext_modules=[dnautils],
    license='LICENSE.txt',
    description='A module for efficient storage and analysis of '
                'high-throughput B-cell sequence data.',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ]
)



import immunedb.api.queries as queries
from immunedb.common.models import CloneStats
from immunedb.util.nested_writer import NestedCSVWriter


class MutationExporter(object):
    def __init__(self, session, clone_ids, limit_sample_ids, thresh_type,
                 thresh_value):
        self._session = session
        self._clone_ids = clone_ids
        self._limit_sample_ids = limit_sample_ids
        self._thresh_type = thresh_type
        self._thresh_value = thresh_value

        headers = ['clone_id', 'sample_id', 'total_seqs', 'region']
        self._numbers = ('unique', 'total')
        self._mtypes = ('synonymous', 'nonsynonymous', 'conservative',
                        'nonconservative', 'unknown')
        for number in self._numbers:
            for mtype in self._mtypes:
                headers.append('{}_{}'.format(mtype, number))

        self._csv = NestedCSVWriter(headers, {
            'nonsynonymous_unique': (
                lambda r: r.get('conservative_unique', 0) +
                r.get('nonconservative_unique', 0)),
            'nonsynonymous_total': (
                lambda r: r.get('conservative_total', 0) +
                r.get('nonconservative_total', 0)),
            'sample_id': lambda r: r['sample_id'] if r['sample_id'] else 'ALL'
        }, streaming=True)

    def _sample_rows(self, cid, sample_id):
        result = queries.get_clone_mutations(
            self._session, cid, self._thresh_type,
            self._thresh_value, sample_id=sample_id)

        for region in sorted(result['regions'].keys()):
            stats = result['regions'][region]
            row = {
                'clone_id': cid,
                'region': region,
                'sample_id': sample_id,
                'total_seqs': result['total_seqs'],
            }

            for number in self._numbers:
                for mtype, count in stats['counts'][number].iteritems():
                    row['{}_{}'.format(mtype, number)] = count
            self._csv.add_row(row, write_default=True, default=0,
                              write_if_stream=False)

    def get_data(self):
        for cid in sorted(self._clone_ids):
            # Write mutations for entire clone
            self._sample_rows(cid, None)
            yield self._csv.get_value()
            # Write per sample mutations
            sample_ids = map(lambda r: r.sample_id, self._session.query(
                CloneStats.sample_id).filter(
                    CloneStats.clone_id == cid,
                    CloneStats.sample_id != 0).all())
            if self._limit_sample_ids is not None:
                sample_ids = set(sample_ids).intersection(
                        set(self._limit_sample_ids))
            for sample_id in sorted(sample_ids):
                self._sample_rows(cid, sample_id)
                yield self._csv.get_value()
